@@ -34,6 +34,15 @@
 #  define _mm256_setr_m128i(v0, v1) _mm256_set_m128i((v1), (v0))
 #endif
 
+#define _MM_CLAMP_PS(r,min,max)				\
+	_mm_min_ps(_mm_max_ps(r, min), max)
+
+#define _MM256_CLAMP_PS(r,min,max)			\
+	_mm256_min_ps(_mm256_max_ps(r, min), max)
+
+#define _MM_CLAMP_SS(r,min,max)				\
+	_mm_min_ss(_mm_max_ss(r, min), max)
+
 static void
 conv_s16_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
@@ -147,7 +156,7 @@ void
 conv_s24_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0];
 	uint32_t n, unrolled;
 	__m128i in;
@@ -164,21 +173,21 @@ conv_s24_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in = _mm_setr_epi32(
 			*((uint32_t*)&s[0 * n_channels]),
-			*((uint32_t*)&s[3 * n_channels]),
-			*((uint32_t*)&s[6 * n_channels]),
-			*((uint32_t*)&s[9 * n_channels]));
+			*((uint32_t*)&s[1 * n_channels]),
+			*((uint32_t*)&s[2 * n_channels]),
+			*((uint32_t*)&s[3 * n_channels]));
 		in = _mm_slli_epi32(in, 8);
 		in = _mm_srai_epi32(in, 8);
 		out = _mm_cvtepi32_ps(in);
 		out = _mm_mul_ps(out, factor);
 		_mm_store_ps(&d0[n], out);
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out = _mm_cvtsi32_ss(factor, read_s24(s));
+		out = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
 		out = _mm_mul_ss(out, factor);
 		_mm_store_ss(&d0[n], out);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 
@@ -186,7 +195,7 @@ static void
 conv_s24_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0], *d1 = dst[1];
 	uint32_t n, unrolled;
 	__m128i in[2];
@@ -205,14 +214,14 @@ conv_s24_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_setr_epi32(
 			*((uint32_t*)&s[0 + 0*n_channels]),
-			*((uint32_t*)&s[0 + 3*n_channels]),
-			*((uint32_t*)&s[0 + 6*n_channels]),
-			*((uint32_t*)&s[0 + 9*n_channels]));
+			*((uint32_t*)&s[0 + 1*n_channels]),
+			*((uint32_t*)&s[0 + 2*n_channels]),
+			*((uint32_t*)&s[0 + 3*n_channels]));
 		in[1] = _mm_setr_epi32(
-			*((uint32_t*)&s[3 + 0*n_channels]),
-			*((uint32_t*)&s[3 + 3*n_channels]),
-			*((uint32_t*)&s[3 + 6*n_channels]),
-			*((uint32_t*)&s[3 + 9*n_channels]));
+			*((uint32_t*)&s[1 + 0*n_channels]),
+			*((uint32_t*)&s[1 + 1*n_channels]),
+			*((uint32_t*)&s[1 + 2*n_channels]),
+			*((uint32_t*)&s[1 + 3*n_channels]));
 
 		in[0] = _mm_slli_epi32(in[0], 8);
 		in[1] = _mm_slli_epi32(in[1], 8);
@@ -229,23 +238,23 @@ conv_s24_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ps(&d0[n], out[0]);
 		_mm_store_ps(&d1[n], out[1]);
 
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out[0] = _mm_cvtsi32_ss(factor, read_s24(s));
-		out[1] = _mm_cvtsi32_ss(factor, read_s24(s+3));
+		out[0] = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
+		out[1] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+1)));
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		_mm_store_ss(&d0[n], out[0]);
 		_mm_store_ss(&d1[n], out[1]);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 static void
 conv_s24_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0], *d1 = dst[1], *d2 = dst[2], *d3 = dst[3];
 	uint32_t n, unrolled;
 	__m128i in[4];
@@ -266,24 +275,24 @@ conv_s24_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_setr_epi32(
 			*((uint32_t*)&s[0 + 0*n_channels]),
-			*((uint32_t*)&s[0 + 3*n_channels]),
-			*((uint32_t*)&s[0 + 6*n_channels]),
-			*((uint32_t*)&s[0 + 9*n_channels]));
+			*((uint32_t*)&s[0 + 1*n_channels]),
+			*((uint32_t*)&s[0 + 2*n_channels]),
+			*((uint32_t*)&s[0 + 3*n_channels]));
 		in[1] = _mm_setr_epi32(
-			*((uint32_t*)&s[3 + 0*n_channels]),
-			*((uint32_t*)&s[3 + 3*n_channels]),
-			*((uint32_t*)&s[3 + 6*n_channels]),
-			*((uint32_t*)&s[3 + 9*n_channels]));
+			*((uint32_t*)&s[1 + 0*n_channels]),
+			*((uint32_t*)&s[1 + 1*n_channels]),
+			*((uint32_t*)&s[1 + 2*n_channels]),
+			*((uint32_t*)&s[1 + 3*n_channels]));
 		in[2] = _mm_setr_epi32(
-			*((uint32_t*)&s[6 + 0*n_channels]),
-			*((uint32_t*)&s[6 + 3*n_channels]),
-			*((uint32_t*)&s[6 + 6*n_channels]),
-			*((uint32_t*)&s[6 + 9*n_channels]));
+			*((uint32_t*)&s[2 + 0*n_channels]),
+			*((uint32_t*)&s[2 + 1*n_channels]),
+			*((uint32_t*)&s[2 + 2*n_channels]),
+			*((uint32_t*)&s[2 + 3*n_channels]));
 		in[3] = _mm_setr_epi32(
-			*((uint32_t*)&s[9 + 0*n_channels]),
-			*((uint32_t*)&s[9 + 3*n_channels]),
-			*((uint32_t*)&s[9 + 6*n_channels]),
-			*((uint32_t*)&s[9 + 9*n_channels]));
+			*((uint32_t*)&s[3 + 0*n_channels]),
+			*((uint32_t*)&s[3 + 1*n_channels]),
+			*((uint32_t*)&s[3 + 2*n_channels]),
+			*((uint32_t*)&s[3 + 3*n_channels]));
 
 		in[0] = _mm_slli_epi32(in[0], 8);
 		in[1] = _mm_slli_epi32(in[1], 8);
@@ -310,13 +319,13 @@ conv_s24_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ps(&d2[n], out[2]);
 		_mm_store_ps(&d3[n], out[3]);
 
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out[0] = _mm_cvtsi32_ss(factor, read_s24(s));
-		out[1] = _mm_cvtsi32_ss(factor, read_s24(s+3));
-		out[2] = _mm_cvtsi32_ss(factor, read_s24(s+6));
-		out[3] = _mm_cvtsi32_ss(factor, read_s24(s+9));
+		out[0] = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
+		out[1] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+1)));
+		out[2] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+2)));
+		out[3] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+3)));
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		out[2] = _mm_mul_ss(out[2], factor);
@@ -325,7 +334,7 @@ conv_s24_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ss(&d1[n], out[1]);
 		_mm_store_ss(&d2[n], out[2]);
 		_mm_store_ss(&d3[n], out[3]);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 
@@ -353,7 +362,7 @@ conv_s32_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	float *d0 = dst[0], *d1 = dst[1], *d2 = dst[2], *d3 = dst[3];
 	uint32_t n, unrolled;
 	__m256i in[4], t[4];
-	__m256 out[4], factor = _mm256_set1_ps(1.0f / S24_SCALE);
+	__m256 out[4], factor = _mm256_set1_ps(1.0f / S32_SCALE);
 	__m256i mask1 = _mm256_setr_epi64x(0*n_channels, 0*n_channels+2, 4*n_channels, 4*n_channels+2);
 	__m256i mask2 = _mm256_setr_epi64x(1*n_channels, 1*n_channels+2, 5*n_channels, 5*n_channels+2);
 	__m256i mask3 = _mm256_setr_epi64x(2*n_channels, 2*n_channels+2, 6*n_channels, 6*n_channels+2);
@@ -373,11 +382,6 @@ conv_s32_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		in[2] = _mm256_i64gather_epi64((long long int *)&s[0*n_channels], mask3, 4);
 		in[3] = _mm256_i64gather_epi64((long long int *)&s[0*n_channels], mask4, 4);
 
-		in[0] = _mm256_srai_epi32(in[0], 8); /* a0 b0 c0 d0 a4 b4 c4 d4 */
-		in[1] = _mm256_srai_epi32(in[1], 8); /* a1 b1 c1 d1 a5 b5 c5 d5 */
-		in[2] = _mm256_srai_epi32(in[2], 8); /* a2 b2 c2 d2 a6 b6 c6 d6 */
-		in[3] = _mm256_srai_epi32(in[3], 8); /* a3 b3 c3 d3 a7 b7 c7 d7 */
-
 		t[0] = _mm256_unpacklo_epi32(in[0], in[1]);   /* a0 a1 b0 b1 a4 a5 b4 b5 */
 		t[1] = _mm256_unpackhi_epi32(in[0], in[1]);   /* c0 c1 d0 d1 c4 c5 d4 d5 */
 		t[2] = _mm256_unpacklo_epi32(in[2], in[3]);   /* a2 a3 b2 b3 a6 a7 b6 b7 */
@@ -405,11 +409,11 @@ conv_s32_to_f32d_4s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		s += 8*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		__m128 out[4], factor = _mm_set1_ps(1.0f / S24_SCALE);
-		out[0] = _mm_cvtsi32_ss(factor, s[0]>>8);
-		out[1] = _mm_cvtsi32_ss(factor, s[1]>>8);
-		out[2] = _mm_cvtsi32_ss(factor, s[2]>>8);
-		out[3] = _mm_cvtsi32_ss(factor, s[3]>>8);
+		__m128 out[4], factor = _mm_set1_ps(1.0f / S32_SCALE);
+		out[0] = _mm_cvtsi32_ss(factor, s[0]);
+		out[1] = _mm_cvtsi32_ss(factor, s[1]);
+		out[2] = _mm_cvtsi32_ss(factor, s[2]);
+		out[3] = _mm_cvtsi32_ss(factor, s[3]);
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		out[2] = _mm_mul_ss(out[2], factor);
@@ -430,7 +434,7 @@ conv_s32_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	float *d0 = dst[0], *d1 = dst[1];
 	uint32_t n, unrolled;
 	__m256i in[4], t[4];
-	__m256 out[4], factor = _mm256_set1_ps(1.0f / S24_SCALE);
+	__m256 out[4], factor = _mm256_set1_ps(1.0f / S32_SCALE);
 	__m256i perm = _mm256_setr_epi32(0, 2, 4, 6, 1, 3, 5, 7);
 	__m256i mask1 = _mm256_setr_epi64x(0*n_channels, 1*n_channels, 2*n_channels, 3*n_channels);
 	__m256i mask2 = _mm256_setr_epi64x(4*n_channels, 5*n_channels, 6*n_channels, 7*n_channels);
@@ -445,9 +449,6 @@ conv_s32_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		in[0] = _mm256_i64gather_epi64((long long int *)s, mask1, 4);
 		in[1] = _mm256_i64gather_epi64((long long int *)s, mask2, 4);
 
-		in[0] = _mm256_srai_epi32(in[0], 8);
-		in[1] = _mm256_srai_epi32(in[1], 8);
-
 		t[0] = _mm256_permutevar8x32_epi32(in[0], perm);
 		t[1] = _mm256_permutevar8x32_epi32(in[1], perm);
 
@@ -466,9 +467,9 @@ conv_s32_to_f32d_2s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		s += 8*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		__m128 out[2], factor = _mm_set1_ps(1.0f / S24_SCALE);
-		out[0] = _mm_cvtsi32_ss(factor, s[0]>>8);
-		out[1] = _mm_cvtsi32_ss(factor, s[1]>>8);
+		__m128 out[2], factor = _mm_set1_ps(1.0f / S32_SCALE);
+		out[0] = _mm_cvtsi32_ss(factor, s[0]);
+		out[1] = _mm_cvtsi32_ss(factor, s[1]);
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		_mm_store_ss(&d0[n], out[0]);
@@ -485,7 +486,7 @@ conv_s32_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	float *d0 = dst[0];
 	uint32_t n, unrolled;
 	__m256i in[2];
-	__m256 out[2], factor = _mm256_set1_ps(1.0f / S24_SCALE);
+	__m256 out[2], factor = _mm256_set1_ps(1.0f / S32_SCALE);
 	__m256i mask1 = _mm256_setr_epi64x(0*n_channels, 1*n_channels, 2*n_channels, 3*n_channels);
 	__m256i mask2 = _mm256_setr_epi64x(4*n_channels, 5*n_channels, 6*n_channels, 7*n_channels);
 
@@ -502,9 +503,6 @@ conv_s32_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 				_mm256_i64gather_epi32(&s[ 8*n_channels], mask1, 4),
 				_mm256_i64gather_epi32(&s[ 8*n_channels], mask2, 4));
 
-		in[0] = _mm256_srai_epi32(in[0], 8);
-		in[1] = _mm256_srai_epi32(in[1], 8);
-
 		out[0] = _mm256_cvtepi32_ps(in[0]);
 		out[1] = _mm256_cvtepi32_ps(in[1]);
 
@@ -517,8 +515,8 @@ conv_s32_to_f32d_1s_avx2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		s += 16*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		__m128 out, factor = _mm_set1_ps(1.0f / S24_SCALE);
-		out = _mm_cvtsi32_ss(factor, s[0]>>8);
+		__m128 out, factor = _mm_set1_ps(1.0f / S32_SCALE);
+		out = _mm_cvtsi32_ss(factor, s[0]);
 		out = _mm_mul_ss(out, factor);
 		_mm_store_ss(&d0[n], out);
 		s += n_channels;
@@ -549,8 +547,9 @@ conv_f32d_to_s32_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[1];
 	__m128i out[4];
-	__m128 scale = _mm_set1_ps(S32_SCALE);
-	__m128 int_max = _mm_set1_ps(S32_MAX);
+	__m128 scale = _mm_set1_ps(S24_SCALE);
+	__m128 int_max = _mm_set1_ps(S24_MAX);
+	__m128 int_min = _mm_set1_ps(S24_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16))
 		unrolled = n_samples & ~3;
@@ -559,8 +558,9 @@ conv_f32d_to_s32_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		out[1] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(0, 3, 2, 1));
 		out[2] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(1, 0, 3, 2));
 		out[3] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(2, 1, 0, 3));
@@ -574,8 +574,8 @@ conv_f32d_to_s32_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	for(; n < n_samples; n++) {
 		in[0] = _mm_load_ss(&s0[n]);
 		in[0] = _mm_mul_ss(in[0], scale);
-		in[0] = _mm_min_ss(in[0], int_max);
-		*d = _mm_cvtss_si32(in[0]);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		*d = _mm_cvtss_si32(in[0]) << 8;
 		d += n_channels;
 	}
 }
@@ -589,8 +589,9 @@ conv_f32d_to_s32_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m256 in[2];
 	__m256i out[2], t[2];
-	__m256 scale = _mm256_set1_ps(S32_SCALE);
-	__m256 int_max = _mm256_set1_ps(S32_MAX);
+	__m256 scale = _mm256_set1_ps(S24_SCALE);
+	__m256 int_min = _mm256_set1_ps(S24_MIN);
+	__m256 int_max = _mm256_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32))
@@ -602,11 +603,13 @@ conv_f32d_to_s32_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n]), scale);
 		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n]), scale);
 
-		in[0] = _mm256_min_ps(in[0], int_max);
-		in[1] = _mm256_min_ps(in[1], int_max);
+		in[0] = _MM256_CLAMP_PS(in[0], int_min, int_max);
+		in[1] = _MM256_CLAMP_PS(in[1], int_min, int_max);
 
-		out[0] = _mm256_cvtps_epi32(in[0]);	/* a0 a1 a2 a3 a4 a5 a6 a7 */
-		out[1] = _mm256_cvtps_epi32(in[1]);	/* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[0] = _mm256_cvttps_epi32(in[0]);	/* a0 a1 a2 a3 a4 a5 a6 a7 */
+		out[1] = _mm256_cvttps_epi32(in[1]);	/* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[0] = _mm256_slli_epi32(out[0], 8);
+		out[1] = _mm256_slli_epi32(out[1], 8);
 
 		t[0] = _mm256_unpacklo_epi32(out[0], out[1]); /* a0 b0 a1 b1 a4 b4 a5 b5 */
 		t[1] = _mm256_unpackhi_epi32(out[0], out[1]); /* a2 b2 a3 b3 a6 b6 a7 b7 */
@@ -635,8 +638,9 @@ conv_f32d_to_s32_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	for(; n < n_samples; n++) {
 		__m128 in[2];
 		__m128i out[2];
-		__m128 scale = _mm_set1_ps(S32_SCALE);
-		__m128 int_max = _mm_set1_ps(S32_MAX);
+		__m128 scale = _mm_set1_ps(S24_SCALE);
+		__m128 int_min = _mm_set1_ps(S24_MIN);
+		__m128 int_max = _mm_set1_ps(S24_MAX);
 
 		in[0] = _mm_load_ss(&s0[n]);
 		in[1] = _mm_load_ss(&s1[n]);
@@ -644,8 +648,9 @@ conv_f32d_to_s32_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm_unpacklo_ps(in[0], in[1]);
 
 		in[0] = _mm_mul_ps(in[0], scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		_mm_storel_epi64((__m128i*)d, out[0]);
 		d += n_channels;
 	}
@@ -660,8 +665,9 @@ conv_f32d_to_s32_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m256 in[4];
 	__m256i out[4], t[4];
-	__m256 scale = _mm256_set1_ps(S32_SCALE);
-	__m256 int_max = _mm256_set1_ps(S32_MAX);
+	__m256 scale = _mm256_set1_ps(S24_SCALE);
+	__m256 int_min = _mm256_set1_ps(S24_MIN);
+	__m256 int_max = _mm256_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32) &&
@@ -677,15 +683,19 @@ conv_f32d_to_s32_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[2] = _mm256_mul_ps(_mm256_load_ps(&s2[n]), scale);
 		in[3] = _mm256_mul_ps(_mm256_load_ps(&s3[n]), scale);
 
-		in[0] = _mm256_min_ps(in[0], int_max);
-		in[1] = _mm256_min_ps(in[1], int_max);
-		in[2] = _mm256_min_ps(in[2], int_max);
-		in[3] = _mm256_min_ps(in[3], int_max);
+		in[0] = _MM256_CLAMP_PS(in[0], int_min, int_max);
+		in[1] = _MM256_CLAMP_PS(in[1], int_min, int_max);
+		in[2] = _MM256_CLAMP_PS(in[2], int_min, int_max);
+		in[3] = _MM256_CLAMP_PS(in[3], int_min, int_max);
 
-		out[0] = _mm256_cvtps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		out[1] = _mm256_cvtps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
-		out[2] = _mm256_cvtps_epi32(in[2]); /* c0 c1 c2 c3 c4 c5 c6 c7 */
-		out[3] = _mm256_cvtps_epi32(in[3]); /* d0 d1 d2 d3 d4 d5 d6 d7 */
+		out[0] = _mm256_cvttps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		out[1] = _mm256_cvttps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[2] = _mm256_cvttps_epi32(in[2]); /* c0 c1 c2 c3 c4 c5 c6 c7 */
+		out[3] = _mm256_cvttps_epi32(in[3]); /* d0 d1 d2 d3 d4 d5 d6 d7 */
+		out[0] = _mm256_slli_epi32(out[0], 8);
+		out[1] = _mm256_slli_epi32(out[1], 8);
+		out[2] = _mm256_slli_epi32(out[2], 8);
+		out[3] = _mm256_slli_epi32(out[3], 8);
 
 		t[0] = _mm256_unpacklo_epi32(out[0], out[1]); /* a0 b0 a1 b1 a4 b4 a5 b5 */
 		t[1] = _mm256_unpackhi_epi32(out[0], out[1]); /* a2 b2 a3 b3 a6 b6 a7 b7 */
@@ -710,8 +720,9 @@ conv_f32d_to_s32_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	for(; n < n_samples; n++) {
 		__m128 in[4];
 		__m128i out[4];
-		__m128 scale = _mm_set1_ps(S32_SCALE);
-		__m128 int_max = _mm_set1_ps(S32_MAX);
+		__m128 scale = _mm_set1_ps(S24_SCALE);
+		__m128 int_min = _mm_set1_ps(S24_MIN);
+		__m128 int_max = _mm_set1_ps(S24_MAX);
 
 		in[0] = _mm_load_ss(&s0[n]);
 		in[1] = _mm_load_ss(&s1[n]);
@@ -723,8 +734,9 @@ conv_f32d_to_s32_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm_unpacklo_ps(in[0], in[1]);
 
 		in[0] = _mm_mul_ps(in[0], scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		_mm_storeu_si128((__m128i*)d, out[0]);
 		d += n_channels;
 	}
@@ -754,8 +766,9 @@ conv_f32d_to_s16_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[2];
 	__m128i out[2];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+	__m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16))
 		unrolled = n_samples & ~7;
@@ -763,10 +776,10 @@ conv_f32d_to_s16_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_scale);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
 		out[0] = _mm_packs_epi32(out[0], out[1]);
 
 		d[0*n_channels] = _mm_extract_epi16(out[0], 0);
@@ -780,8 +793,8 @@ conv_f32d_to_s16_1s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		d += 8*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
 		*d = _mm_cvtss_si32(in[0]);
 		d += n_channels;
 	}
@@ -796,7 +809,7 @@ conv_f32d_to_s16_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m256 in[2];
 	__m256i out[4], t[2];
-	__m256 int_max = _mm256_set1_ps(S16_MAX_F);
+	__m256 int_scale = _mm256_set1_ps(S16_SCALE);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32))
@@ -805,11 +818,11 @@ conv_f32d_to_s16_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n+0]), int_max);
-		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n+0]), int_max);
+		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n+0]), int_scale);
+		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n+0]), int_scale);
 
-		out[0] = _mm256_cvtps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		out[1] = _mm256_cvtps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[0] = _mm256_cvttps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		out[1] = _mm256_cvttps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
 
 		t[0] = _mm256_unpacklo_epi32(out[0], out[1]); /* a0 b0 a1 b1 a4 b4 a5 b5 */
 		t[1] = _mm256_unpackhi_epi32(out[0], out[1]); /* a2 b2 a3 b3 a6 b6 a7 b7 */
@@ -829,13 +842,14 @@ conv_f32d_to_s16_2s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	}
 	for(; n < n_samples; n++) {
 		__m128 in[2];
-		__m128 int_max = _mm_set1_ps(S16_MAX_F);
-	        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+		__m128 int_scale = _mm_set1_ps(S16_SCALE);
+		__m128 int_max = _mm_set1_ps(S16_MAX);
+		__m128 int_min = _mm_set1_ps(S16_MIN);
 
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d += n_channels;
@@ -851,7 +865,7 @@ conv_f32d_to_s16_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m256 in[4];
 	__m256i out[4], t[4];
-	__m256 int_max = _mm256_set1_ps(S16_MAX_F);
+	__m256 int_scale = _mm256_set1_ps(S16_SCALE);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32) &&
@@ -862,15 +876,15 @@ conv_f32d_to_s16_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n]), int_max);
-		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n]), int_max);
-		in[2] = _mm256_mul_ps(_mm256_load_ps(&s2[n]), int_max);
-		in[3] = _mm256_mul_ps(_mm256_load_ps(&s3[n]), int_max);
+		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n]), int_scale);
+		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n]), int_scale);
+		in[2] = _mm256_mul_ps(_mm256_load_ps(&s2[n]), int_scale);
+		in[3] = _mm256_mul_ps(_mm256_load_ps(&s3[n]), int_scale);
 
-		t[0] = _mm256_cvtps_epi32(in[0]);  /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		t[1] = _mm256_cvtps_epi32(in[1]);  /* b0 b1 b2 b3 b4 b5 b6 b7 */
-		t[2] = _mm256_cvtps_epi32(in[2]);  /* c0 c1 c2 c3 c4 c5 c6 c7 */
-		t[3] = _mm256_cvtps_epi32(in[3]);  /* d0 d1 d2 d3 d4 d5 d6 d7 */
+		t[0] = _mm256_cvttps_epi32(in[0]);  /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		t[1] = _mm256_cvttps_epi32(in[1]);  /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		t[2] = _mm256_cvttps_epi32(in[2]);  /* c0 c1 c2 c3 c4 c5 c6 c7 */
+		t[3] = _mm256_cvttps_epi32(in[3]);  /* d0 d1 d2 d3 d4 d5 d6 d7 */
 
 		t[0] = _mm256_packs_epi32(t[0], t[2]); /* a0 a1 a2 a3 c0 c1 c2 c3 a4 a5 a6 a7 c4 c5 c6 c7 */
 		t[1] = _mm256_packs_epi32(t[1], t[3]); /* b0 b1 b2 b3 d0 d1 d2 d3 b4 b5 b6 b7 d4 d5 d6 d7 */
@@ -905,17 +919,18 @@ conv_f32d_to_s16_4s_avx2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	}
 	for(; n < n_samples; n++) {
 		__m128 in[4];
-		__m128 int_max = _mm_set1_ps(S16_MAX_F);
-	        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+		__m128 int_scale = _mm_set1_ps(S16_SCALE);
+		__m128 int_max = _mm_set1_ps(S16_MAX);
+		__m128 int_min = _mm_set1_ps(S16_MIN);
 
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_max);
-		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
-		in[2] = _mm_min_ss(int_max, _mm_max_ss(in[2], int_min));
-		in[3] = _mm_min_ss(int_max, _mm_max_ss(in[3], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_scale);
+		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
+		in[2] = _MM_CLAMP_SS(in[2], int_min, int_max);
+		in[3] = _MM_CLAMP_SS(in[3], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d[2] = _mm_cvtss_si32(in[2]);
@@ -948,7 +963,7 @@ conv_f32d_to_s16_4_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 	uint32_t n, unrolled;
 	__m256 in[4];
 	__m256i out[4], t[4];
-	__m256 int_max = _mm256_set1_ps(S16_MAX_F);
+	__m256 int_scale = _mm256_set1_ps(S16_SCALE);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32) &&
@@ -959,15 +974,15 @@ conv_f32d_to_s16_4_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n]), int_max);
-		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n]), int_max);
-		in[2] = _mm256_mul_ps(_mm256_load_ps(&s2[n]), int_max);
-		in[3] = _mm256_mul_ps(_mm256_load_ps(&s3[n]), int_max);
+		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n]), int_scale);
+		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n]), int_scale);
+		in[2] = _mm256_mul_ps(_mm256_load_ps(&s2[n]), int_scale);
+		in[3] = _mm256_mul_ps(_mm256_load_ps(&s3[n]), int_scale);
 
-		t[0] = _mm256_cvtps_epi32(in[0]);  /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		t[1] = _mm256_cvtps_epi32(in[1]);  /* b0 b1 b2 b3 b4 b5 b6 b7 */
-		t[2] = _mm256_cvtps_epi32(in[2]);  /* c0 c1 c2 c3 c4 c5 c6 c7 */
-		t[3] = _mm256_cvtps_epi32(in[3]);  /* d0 d1 d2 d3 d4 d5 d6 d7 */
+		t[0] = _mm256_cvttps_epi32(in[0]);  /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		t[1] = _mm256_cvttps_epi32(in[1]);  /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		t[2] = _mm256_cvttps_epi32(in[2]);  /* c0 c1 c2 c3 c4 c5 c6 c7 */
+		t[3] = _mm256_cvttps_epi32(in[3]);  /* d0 d1 d2 d3 d4 d5 d6 d7 */
 
 		t[0] = _mm256_packs_epi32(t[0], t[2]); /* a0 a1 a2 a3 c0 c1 c2 c3 a4 a5 a6 a7 c4 c5 c6 c7 */
 		t[1] = _mm256_packs_epi32(t[1], t[3]); /* b0 b1 b2 b3 d0 d1 d2 d3 b4 b5 b6 b7 d4 d5 d6 d7 */
@@ -987,17 +1002,18 @@ conv_f32d_to_s16_4_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 	}
 	for(; n < n_samples; n++) {
 		__m128 in[4];
-		__m128 int_max = _mm_set1_ps(S16_MAX_F);
-	        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+		__m128 int_scale = _mm_set1_ps(S16_SCALE);
+		__m128 int_max = _mm_set1_ps(S16_MAX);
+		__m128 int_min = _mm_set1_ps(S16_MIN);
 
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_max);
-		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
-		in[2] = _mm_min_ss(int_max, _mm_max_ss(in[2], int_min));
-		in[3] = _mm_min_ss(int_max, _mm_max_ss(in[3], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_scale);
+		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
+		in[2] = _MM_CLAMP_SS(in[2], int_min, int_max);
+		in[3] = _MM_CLAMP_SS(in[3], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d[2] = _mm_cvtss_si32(in[2]);
@@ -1014,7 +1030,7 @@ conv_f32d_to_s16_2_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 	uint32_t n, unrolled;
 	__m256 in[4];
 	__m256i out[4], t[4];
-	__m256 int_max = _mm256_set1_ps(S16_MAX_F);
+	__m256 int_scale = _mm256_set1_ps(S16_SCALE);
 
 	if (SPA_IS_ALIGNED(s0, 32) &&
 	    SPA_IS_ALIGNED(s1, 32))
@@ -1023,15 +1039,15 @@ conv_f32d_to_s16_2_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 16) {
-		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n+0]), int_max);
-		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n+0]), int_max);
-		in[2] = _mm256_mul_ps(_mm256_load_ps(&s0[n+8]), int_max);
-		in[3] = _mm256_mul_ps(_mm256_load_ps(&s1[n+8]), int_max);
+		in[0] = _mm256_mul_ps(_mm256_load_ps(&s0[n+0]), int_scale);
+		in[1] = _mm256_mul_ps(_mm256_load_ps(&s1[n+0]), int_scale);
+		in[2] = _mm256_mul_ps(_mm256_load_ps(&s0[n+8]), int_scale);
+		in[3] = _mm256_mul_ps(_mm256_load_ps(&s1[n+8]), int_scale);
 
-		out[0] = _mm256_cvtps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		out[1] = _mm256_cvtps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
-		out[2] = _mm256_cvtps_epi32(in[2]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
-		out[3] = _mm256_cvtps_epi32(in[3]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[0] = _mm256_cvttps_epi32(in[0]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		out[1] = _mm256_cvttps_epi32(in[1]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
+		out[2] = _mm256_cvttps_epi32(in[2]); /* a0 a1 a2 a3 a4 a5 a6 a7 */
+		out[3] = _mm256_cvttps_epi32(in[3]); /* b0 b1 b2 b3 b4 b5 b6 b7 */
 
 		t[0] = _mm256_unpacklo_epi32(out[0], out[1]); /* a0 b0 a1 b1 a4 b4 a5 b5 */
 		t[1] = _mm256_unpackhi_epi32(out[0], out[1]); /* a2 b2 a3 b3 a6 b6 a7 b7 */
@@ -1048,13 +1064,14 @@ conv_f32d_to_s16_2_avx2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 	}
 	for(; n < n_samples; n++) {
 		__m128 in[4];
-		__m128 int_max = _mm_set1_ps(S16_MAX_F);
-		__m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+		__m128 int_scale = _mm_set1_ps(S16_SCALE);
+		__m128 int_max = _mm_set1_ps(S16_MAX);
+		__m128 int_min = _mm_set1_ps(S16_MIN);
 
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d += 2;
