@@ -26,6 +26,12 @@
 
 #include <emmintrin.h>
 
+#define _MM_CLAMP_PS(r,min,max)				\
+	_mm_min_ps(_mm_max_ps(r, min), max)
+
+#define _MM_CLAMP_SS(r,min,max)				\
+	_mm_min_ss(_mm_max_ss(r, min), max)
+
 static void
 conv_s16_to_f32d_1s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
@@ -132,7 +138,7 @@ void
 conv_s24_to_f32d_1s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0];
 	uint32_t n, unrolled;
 	__m128i in;
@@ -149,21 +155,21 @@ conv_s24_to_f32d_1s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in = _mm_setr_epi32(
 			*((uint32_t*)&s[0 * n_channels]),
-			*((uint32_t*)&s[3 * n_channels]),
-			*((uint32_t*)&s[6 * n_channels]),
-			*((uint32_t*)&s[9 * n_channels]));
+			*((uint32_t*)&s[1 * n_channels]),
+			*((uint32_t*)&s[2 * n_channels]),
+			*((uint32_t*)&s[3 * n_channels]));
 		in = _mm_slli_epi32(in, 8);
 		in = _mm_srai_epi32(in, 8);
 		out = _mm_cvtepi32_ps(in);
 		out = _mm_mul_ps(out, factor);
 		_mm_store_ps(&d0[n], out);
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out = _mm_cvtsi32_ss(factor, read_s24(s));
+		out = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
 		out = _mm_mul_ss(out, factor);
 		_mm_store_ss(&d0[n], out);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 
@@ -171,7 +177,7 @@ static void
 conv_s24_to_f32d_2s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0], *d1 = dst[1];
 	uint32_t n, unrolled;
 	__m128i in[2];
@@ -190,14 +196,14 @@ conv_s24_to_f32d_2s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_setr_epi32(
 			*((uint32_t*)&s[0 + 0*n_channels]),
-			*((uint32_t*)&s[0 + 3*n_channels]),
-			*((uint32_t*)&s[0 + 6*n_channels]),
-			*((uint32_t*)&s[0 + 9*n_channels]));
+			*((uint32_t*)&s[0 + 1*n_channels]),
+			*((uint32_t*)&s[0 + 2*n_channels]),
+			*((uint32_t*)&s[0 + 3*n_channels]));
 		in[1] = _mm_setr_epi32(
-			*((uint32_t*)&s[3 + 0*n_channels]),
-			*((uint32_t*)&s[3 + 3*n_channels]),
-			*((uint32_t*)&s[3 + 6*n_channels]),
-			*((uint32_t*)&s[3 + 9*n_channels]));
+			*((uint32_t*)&s[1 + 0*n_channels]),
+			*((uint32_t*)&s[1 + 1*n_channels]),
+			*((uint32_t*)&s[1 + 2*n_channels]),
+			*((uint32_t*)&s[1 + 3*n_channels]));
 
 		in[0] = _mm_slli_epi32(in[0], 8);
 		in[1] = _mm_slli_epi32(in[1], 8);
@@ -214,23 +220,23 @@ conv_s24_to_f32d_2s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ps(&d0[n], out[0]);
 		_mm_store_ps(&d1[n], out[1]);
 
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out[0] = _mm_cvtsi32_ss(factor, read_s24(s));
-		out[1] = _mm_cvtsi32_ss(factor, read_s24(s+3));
+		out[0] = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
+		out[1] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+1)));
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		_mm_store_ss(&d0[n], out[0]);
 		_mm_store_ss(&d1[n], out[1]);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 static void
 conv_s24_to_f32d_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src,
 		uint32_t n_channels, uint32_t n_samples)
 {
-	const uint8_t *s = src;
+	const int24_t *s = src;
 	float *d0 = dst[0], *d1 = dst[1], *d2 = dst[2], *d3 = dst[3];
 	uint32_t n, unrolled;
 	__m128i in[4];
@@ -251,24 +257,24 @@ conv_s24_to_f32d_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_setr_epi32(
 			*((uint32_t*)&s[0 + 0*n_channels]),
-			*((uint32_t*)&s[0 + 3*n_channels]),
-			*((uint32_t*)&s[0 + 6*n_channels]),
-			*((uint32_t*)&s[0 + 9*n_channels]));
+			*((uint32_t*)&s[0 + 1*n_channels]),
+			*((uint32_t*)&s[0 + 2*n_channels]),
+			*((uint32_t*)&s[0 + 3*n_channels]));
 		in[1] = _mm_setr_epi32(
-			*((uint32_t*)&s[3 + 0*n_channels]),
-			*((uint32_t*)&s[3 + 3*n_channels]),
-			*((uint32_t*)&s[3 + 6*n_channels]),
-			*((uint32_t*)&s[3 + 9*n_channels]));
+			*((uint32_t*)&s[1 + 0*n_channels]),
+			*((uint32_t*)&s[1 + 1*n_channels]),
+			*((uint32_t*)&s[1 + 2*n_channels]),
+			*((uint32_t*)&s[1 + 3*n_channels]));
 		in[2] = _mm_setr_epi32(
-			*((uint32_t*)&s[6 + 0*n_channels]),
-			*((uint32_t*)&s[6 + 3*n_channels]),
-			*((uint32_t*)&s[6 + 6*n_channels]),
-			*((uint32_t*)&s[6 + 9*n_channels]));
+			*((uint32_t*)&s[2 + 0*n_channels]),
+			*((uint32_t*)&s[2 + 1*n_channels]),
+			*((uint32_t*)&s[2 + 2*n_channels]),
+			*((uint32_t*)&s[2 + 3*n_channels]));
 		in[3] = _mm_setr_epi32(
-			*((uint32_t*)&s[9 + 0*n_channels]),
-			*((uint32_t*)&s[9 + 3*n_channels]),
-			*((uint32_t*)&s[9 + 6*n_channels]),
-			*((uint32_t*)&s[9 + 9*n_channels]));
+			*((uint32_t*)&s[3 + 0*n_channels]),
+			*((uint32_t*)&s[3 + 1*n_channels]),
+			*((uint32_t*)&s[3 + 2*n_channels]),
+			*((uint32_t*)&s[3 + 3*n_channels]));
 
 		in[0] = _mm_slli_epi32(in[0], 8);
 		in[1] = _mm_slli_epi32(in[1], 8);
@@ -295,13 +301,13 @@ conv_s24_to_f32d_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ps(&d2[n], out[2]);
 		_mm_store_ps(&d3[n], out[3]);
 
-		s += 12 * n_channels;
+		s += 4 * n_channels;
 	}
 	for(; n < n_samples; n++) {
-		out[0] = _mm_cvtsi32_ss(factor, read_s24(s));
-		out[1] = _mm_cvtsi32_ss(factor, read_s24(s+3));
-		out[2] = _mm_cvtsi32_ss(factor, read_s24(s+6));
-		out[3] = _mm_cvtsi32_ss(factor, read_s24(s+9));
+		out[0] = _mm_cvtsi32_ss(factor, s24_to_s32(*s));
+		out[1] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+1)));
+		out[2] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+2)));
+		out[3] = _mm_cvtsi32_ss(factor, s24_to_s32(*(s+3)));
 		out[0] = _mm_mul_ss(out[0], factor);
 		out[1] = _mm_mul_ss(out[1], factor);
 		out[2] = _mm_mul_ss(out[2], factor);
@@ -310,7 +316,7 @@ conv_s24_to_f32d_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void * SPA
 		_mm_store_ss(&d1[n], out[1]);
 		_mm_store_ss(&d2[n], out[2]);
 		_mm_store_ss(&d3[n], out[3]);
-		s += 3 * n_channels;
+		s += n_channels;
 	}
 }
 
@@ -384,8 +390,9 @@ conv_f32d_to_s32_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[1];
 	__m128i out[4];
-	__m128 scale = _mm_set1_ps(S32_SCALE);
-	__m128 int_max = _mm_set1_ps(S32_MAX);
+	__m128 scale = _mm_set1_ps(S24_SCALE);
+	__m128 int_min = _mm_set1_ps(S24_MIN);
+	__m128 int_max = _mm_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s0, 16))
 		unrolled = n_samples & ~3;
@@ -394,8 +401,9 @@ conv_f32d_to_s32_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		out[1] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(0, 3, 2, 1));
 		out[2] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(1, 0, 3, 2));
 		out[3] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(2, 1, 0, 3));
@@ -409,8 +417,8 @@ conv_f32d_to_s32_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	for(; n < n_samples; n++) {
 		in[0] = _mm_load_ss(&s0[n]);
 		in[0] = _mm_mul_ss(in[0], scale);
-		in[0] = _mm_min_ss(in[0], int_max);
-		*d = _mm_cvtss_si32(in[0]);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		*d = _mm_cvtss_si32(in[0]) << 8;
 		d += n_channels;
 	}
 }
@@ -424,8 +432,9 @@ conv_f32d_to_s32_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[2];
 	__m128i out[2], t[2];
-	__m128 scale = _mm_set1_ps(S32_SCALE);
-	__m128 int_max = _mm_set1_ps(S32_MAX);
+	__m128 scale = _mm_set1_ps(S24_SCALE);
+	__m128 int_min = _mm_set1_ps(S24_MIN);
+	__m128 int_max = _mm_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s0, 16) &&
 	    SPA_IS_ALIGNED(s1, 16))
@@ -437,11 +446,13 @@ conv_f32d_to_s32_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), scale);
 		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n]), scale);
 
-		in[0] = _mm_min_ps(in[0], int_max);
-		in[1] = _mm_min_ps(in[1], int_max);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_PS(in[1], int_min, int_max);
 
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
+		out[0] = _mm_slli_epi32(out[0], 8);
+		out[1] = _mm_slli_epi32(out[1], 8);
 
 		t[0] = _mm_unpacklo_epi32(out[0], out[1]);
 		t[1] = _mm_unpackhi_epi32(out[0], out[1]);
@@ -459,8 +470,9 @@ conv_f32d_to_s32_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm_unpacklo_ps(in[0], in[1]);
 
 		in[0] = _mm_mul_ps(in[0], scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		_mm_storel_epi64((__m128i*)d, out[0]);
 		d += n_channels;
 	}
@@ -475,8 +487,9 @@ conv_f32d_to_s32_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[4];
 	__m128i out[4];
-	__m128 scale = _mm_set1_ps(S32_SCALE);
-	__m128 int_max = _mm_set1_ps(S32_MAX);
+	__m128 scale = _mm_set1_ps(S24_SCALE);
+	__m128 int_min = _mm_set1_ps(S24_MIN);
+	__m128 int_max = _mm_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s0, 16) &&
 	    SPA_IS_ALIGNED(s1, 16) &&
@@ -492,17 +505,21 @@ conv_f32d_to_s32_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[2] = _mm_mul_ps(_mm_load_ps(&s2[n]), scale);
 		in[3] = _mm_mul_ps(_mm_load_ps(&s3[n]), scale);
 
-		in[0] = _mm_min_ps(in[0], int_max);
-		in[1] = _mm_min_ps(in[1], int_max);
-		in[2] = _mm_min_ps(in[2], int_max);
-		in[3] = _mm_min_ps(in[3], int_max);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_PS(in[1], int_min, int_max);
+		in[2] = _MM_CLAMP_PS(in[2], int_min, int_max);
+		in[3] = _MM_CLAMP_PS(in[3], int_min, int_max);
 
 		_MM_TRANSPOSE4_PS(in[0], in[1], in[2], in[3]);
 
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
-		out[2] = _mm_cvtps_epi32(in[2]);
-		out[3] = _mm_cvtps_epi32(in[3]);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
+		out[2] = _mm_cvttps_epi32(in[2]);
+		out[3] = _mm_cvttps_epi32(in[3]);
+		out[0] = _mm_slli_epi32(out[0], 8);
+		out[1] = _mm_slli_epi32(out[1], 8);
+		out[2] = _mm_slli_epi32(out[2], 8);
+		out[3] = _mm_slli_epi32(out[3], 8);
 
 		_mm_storeu_si128((__m128i*)(d + 0*n_channels), out[0]);
 		_mm_storeu_si128((__m128i*)(d + 1*n_channels), out[1]);
@@ -521,8 +538,9 @@ conv_f32d_to_s32_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		in[0] = _mm_unpacklo_ps(in[0], in[1]);
 
 		in[0] = _mm_mul_ps(in[0], scale);
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		_mm_storeu_si128((__m128i*)d, out[0]);
 		d += n_channels;
 	}
@@ -543,28 +561,45 @@ conv_f32d_to_s32_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const voi
 		conv_f32d_to_s32_1s_sse2(conv, &d[i], &src[i], n_channels, n_samples);
 }
 
+/* 32 bit xorshift PRNG, see https://en.wikipedia.org/wiki/Xorshift */
+#define _MM_XORSHIFT_EPI32(r)				\
+({							\
+	__m128i i, t;					\
+	i = _mm_load_si128((__m128i*)r);		\
+	t = _mm_slli_epi32(i, 13);			\
+	i = _mm_xor_si128(i, t);			\
+	t = _mm_srli_epi32(i, 17);			\
+	i = _mm_xor_si128(i, t);			\
+	t = _mm_slli_epi32(i, 5);			\
+	i = _mm_xor_si128(i, t);			\
+	_mm_store_si128((__m128i*)r, i);		\
+	i;						\
+})
+
+
 static inline void update_dither_sse2(struct convert *conv, uint32_t n_samples)
 {
 	uint32_t n;
 	const uint32_t *r = SPA_PTR_ALIGN(conv->random, 16, uint32_t);
+	__m128 scale = _mm_set1_ps(conv->scale);
+	__m128 out[1];
 	float *dither = SPA_PTR_ALIGN(conv->dither, 16, float);
-	__m128 scale = _mm_set1_ps(conv->scale), out[1];
-	__m128i in[1], t[1];
+	__m128i in[2];
 
-	for (n = 0; n < n_samples; n += 4) {
-		/* 32 bit xorshift PRNG, see https://en.wikipedia.org/wiki/Xorshift */
-		in[0] = _mm_load_si128((__m128i*)r);
-		t[0] = _mm_slli_epi32(in[0], 13);
-		in[0] = _mm_xor_si128(in[0], t[0]);
-		t[0] = _mm_srli_epi32(in[0], 17);
-		in[0] = _mm_xor_si128(in[0], t[0]);
-		t[0] = _mm_slli_epi32(in[0], 5);
-		in[0] = _mm_xor_si128(in[0], t[0]);
-		_mm_store_si128((__m128i*)r, in[0]);
-
-		out[0] = _mm_cvtepi32_ps(in[0]);
-		out[0] = _mm_mul_ps(out[0], scale);
-		_mm_store_ps(&dither[n], out[0]);
+	if (conv->method < DITHER_METHOD_TRIANGULAR) {
+		for (n = 0; n < n_samples; n += 4) {
+			in[0] = _MM_XORSHIFT_EPI32(r);
+			out[0] = _mm_cvtepi32_ps(_MM_XORSHIFT_EPI32(r));
+			out[0] = _mm_mul_ps(out[0], scale);
+			_mm_store_ps(&dither[n], out[0]);
+		}
+	} else {
+		for (n = 0; n < n_samples; n += 4) {
+			in[0] = _mm_add_epi32( _MM_XORSHIFT_EPI32(r), _MM_XORSHIFT_EPI32(r));
+			out[0] = _mm_cvtepi32_ps(in[0]);
+			out[0] = _mm_mul_ps(out[0], scale);
+			_mm_store_ps(&dither[n], out[0]);
+		}
 	}
 }
 
@@ -578,8 +613,9 @@ conv_f32d_to_s32_1s_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst, c
 	uint32_t n, unrolled;
 	__m128 in[1];
 	__m128i out[4];
-	__m128 scale = _mm_set1_ps(S32_SCALE);
-	__m128 int_max = _mm_set1_ps(S32_MAX);
+	__m128 scale = _mm_set1_ps(S24_SCALE);
+	__m128 int_min = _mm_set1_ps(S24_MIN);
+	__m128 int_max = _mm_set1_ps(S24_MAX);
 
 	if (SPA_IS_ALIGNED(s, 16))
 		unrolled = n_samples & ~3;
@@ -589,8 +625,9 @@ conv_f32d_to_s32_1s_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst, c
 	for(n = 0; n < unrolled; n += 4) {
 		in[0] = _mm_mul_ps(_mm_load_ps(&s[n]), scale);
 		in[0] = _mm_add_ps(in[0], _mm_load_ps(&dither[n]));
-		in[0] = _mm_min_ps(in[0], int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
+		in[0] = _MM_CLAMP_PS(in[0], int_min, int_max);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[0] = _mm_slli_epi32(out[0], 8);
 		out[1] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(0, 3, 2, 1));
 		out[2] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(1, 0, 3, 2));
 		out[3] = _mm_shuffle_epi32(out[0], _MM_SHUFFLE(2, 1, 0, 3));
@@ -605,8 +642,8 @@ conv_f32d_to_s32_1s_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst, c
 		in[0] = _mm_load_ss(&s[n]);
 		in[0] = _mm_mul_ss(in[0], scale);
 		in[0] = _mm_add_ss(in[0], _mm_load_ss(&dither[n]));
-		in[0] = _mm_min_ss(in[0], int_max);
-		*d = _mm_cvtss_si32(in[0]);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		*d = _mm_cvtss_si32(in[0]) << 8;
 		d += n_channels;
 	}
 }
@@ -699,7 +736,7 @@ conv_interleave_32_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA
 }
 
 void
-conv_interleave_32_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+conv_32d_to_32_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 		uint32_t n_samples)
 {
 	int32_t *d = dst[0];
@@ -797,7 +834,7 @@ conv_interleave_32s_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SP
 }
 
 void
-conv_interleave_32s_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+conv_32d_to_32s_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 		uint32_t n_samples)
 {
 	int32_t *d = dst[0];
@@ -878,7 +915,7 @@ conv_deinterleave_32_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void *
 }
 
 void
-conv_deinterleave_32_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+conv_32_to_32d_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 		uint32_t n_samples)
 {
 	const float *s = src[0];
@@ -965,7 +1002,7 @@ conv_deinterleave_32s_4s_sse2(void *data, void * SPA_RESTRICT dst[], const void
 }
 
 void
-conv_deinterleave_32s_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+conv_32s_to_32d_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 		uint32_t n_samples)
 {
 	const float *s = src[0];
@@ -986,8 +1023,9 @@ conv_f32_to_s16_1_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_RES
 	uint32_t n, unrolled;
 	__m128 in[2];
 	__m128i out[2];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s, 16))
 		unrolled = n_samples & ~7;
@@ -995,17 +1033,17 @@ conv_f32_to_s16_1_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_RES
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s[n]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s[n+4]), int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s[n+4]), int_scale);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
 		out[0] = _mm_packs_epi32(out[0], out[1]);
 		_mm_storeu_si128((__m128i*)(d+0), out[0]);
 		d += 8;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
 		*d++ = _mm_cvtss_si32(in[0]);
 	}
 }
@@ -1035,8 +1073,9 @@ conv_f32d_to_s16_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[2];
 	__m128i out[2];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16))
 		unrolled = n_samples & ~7;
@@ -1044,10 +1083,10 @@ conv_f32d_to_s16_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_max);
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_scale);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
 		out[0] = _mm_packs_epi32(out[0], out[1]);
 
 		d[0*n_channels] = _mm_extract_epi16(out[0], 0);
@@ -1061,9 +1100,9 @@ conv_f32d_to_s16_1s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		d += 8*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		*d = _mm_cvtss_si32(in[0]);
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		*d = _mm_cvttss_si32(in[0]);
 		d += n_channels;
 	}
 }
@@ -1077,8 +1116,9 @@ conv_f32d_to_s16_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[2];
 	__m128i out[4], t[2];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16) &&
 	    SPA_IS_ALIGNED(s1, 16))
@@ -1087,11 +1127,11 @@ conv_f32d_to_s16_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 4) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n]), int_max);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n]), int_scale);
 
-		t[0] = _mm_cvtps_epi32(in[0]);
-		t[1] = _mm_cvtps_epi32(in[1]);
+		t[0] = _mm_cvttps_epi32(in[0]);
+		t[1] = _mm_cvttps_epi32(in[1]);
 
 		t[0] = _mm_packs_epi32(t[0], t[0]);
 		t[1] = _mm_packs_epi32(t[1], t[1]);
@@ -1108,10 +1148,10 @@ conv_f32d_to_s16_2s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		d += 4*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d += n_channels;
@@ -1127,8 +1167,9 @@ conv_f32d_to_s16_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 	uint32_t n, unrolled;
 	__m128 in[4];
 	__m128i out[4], t[4];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16) &&
 	    SPA_IS_ALIGNED(s1, 16) &&
@@ -1139,15 +1180,15 @@ conv_f32d_to_s16_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 4) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n]), int_max);
-		in[2] = _mm_mul_ps(_mm_load_ps(&s2[n]), int_max);
-		in[3] = _mm_mul_ps(_mm_load_ps(&s3[n]), int_max);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n]), int_scale);
+		in[2] = _mm_mul_ps(_mm_load_ps(&s2[n]), int_scale);
+		in[3] = _mm_mul_ps(_mm_load_ps(&s3[n]), int_scale);
 
-		t[0] = _mm_cvtps_epi32(in[0]);
-		t[1] = _mm_cvtps_epi32(in[1]);
-		t[2] = _mm_cvtps_epi32(in[2]);
-		t[3] = _mm_cvtps_epi32(in[3]);
+		t[0] = _mm_cvttps_epi32(in[0]);
+		t[1] = _mm_cvttps_epi32(in[1]);
+		t[2] = _mm_cvttps_epi32(in[2]);
+		t[3] = _mm_cvttps_epi32(in[3]);
 
 		t[0] = _mm_packs_epi32(t[0], t[2]);
 		t[1] = _mm_packs_epi32(t[1], t[3]);
@@ -1165,14 +1206,14 @@ conv_f32d_to_s16_4s_sse2(void *data, void * SPA_RESTRICT dst, const void * SPA_R
 		d += 4*n_channels;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_max);
-		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
-		in[2] = _mm_min_ss(int_max, _mm_max_ss(in[2], int_min));
-		in[3] = _mm_min_ss(int_max, _mm_max_ss(in[3], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[2] = _mm_mul_ss(_mm_load_ss(&s2[n]), int_scale);
+		in[3] = _mm_mul_ss(_mm_load_ss(&s3[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
+		in[2] = _MM_CLAMP_SS(in[2], int_min, int_max);
+		in[3] = _MM_CLAMP_SS(in[3], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d[2] = _mm_cvtss_si32(in[2]);
@@ -1196,6 +1237,126 @@ conv_f32d_to_s16_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const voi
 		conv_f32d_to_s16_1s_sse2(conv, &d[i], &src[i], n_channels, n_samples);
 }
 
+static void
+conv_f32d_to_s16_1s_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst, const void * SPA_RESTRICT src,
+		uint32_t n_channels, uint32_t n_samples)
+{
+	const float *s0 = src;
+	int16_t *d = dst;
+	float *dither = SPA_PTR_ALIGN(conv->dither, 16, float);
+	uint32_t n, unrolled;
+	__m128 in[2];
+	__m128i out[2];
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
+
+	if (SPA_IS_ALIGNED(s0, 16))
+		unrolled = n_samples & ~7;
+	else
+		unrolled = 0;
+
+	for(n = 0; n < unrolled; n += 8) {
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_scale);
+		in[0] = _mm_add_ps(in[0], _mm_load_ps(&dither[n]));
+		in[1] = _mm_add_ps(in[1], _mm_load_ps(&dither[n+4]));
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
+		out[0] = _mm_packs_epi32(out[0], out[1]);
+
+		d[0*n_channels] = _mm_extract_epi16(out[0], 0);
+		d[1*n_channels] = _mm_extract_epi16(out[0], 1);
+		d[2*n_channels] = _mm_extract_epi16(out[0], 2);
+		d[3*n_channels] = _mm_extract_epi16(out[0], 3);
+		d[4*n_channels] = _mm_extract_epi16(out[0], 4);
+		d[5*n_channels] = _mm_extract_epi16(out[0], 5);
+		d[6*n_channels] = _mm_extract_epi16(out[0], 6);
+		d[7*n_channels] = _mm_extract_epi16(out[0], 7);
+		d += 8*n_channels;
+	}
+	for(; n < n_samples; n++) {
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[0] = _mm_add_ss(in[0], _mm_load_ss(&dither[n]));
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		*d = _mm_cvttss_si32(in[0]);
+		d += n_channels;
+	}
+}
+
+void
+conv_f32d_to_s16_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+		uint32_t n_samples)
+{
+	int16_t *d = dst[0];
+	uint32_t i, k, chunk, n_channels = conv->n_channels;
+
+	update_dither_sse2(conv, SPA_MIN(n_samples, conv->dither_size));
+
+	for(i = 0; i < n_channels; i++) {
+		const float *s = src[i];
+		for(k = 0; k < n_samples; k += chunk) {
+			chunk = SPA_MIN(n_samples - k, conv->dither_size);
+			conv_f32d_to_s16_1s_dither_sse2(conv, &d[i + k*n_channels], &s[k], n_channels, chunk);
+		}
+	}
+}
+
+static void
+conv_f32_to_s16_1_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst, const void * SPA_RESTRICT src,
+		uint32_t n_samples)
+{
+	const float *s = src;
+	int16_t *d = dst;
+	float *dither = SPA_PTR_ALIGN(conv->dither, 16, float);
+	uint32_t n, unrolled;
+	__m128 in[2];
+	__m128i out[2];
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
+
+	if (SPA_IS_ALIGNED(s, 16))
+		unrolled = n_samples & ~7;
+	else
+		unrolled = 0;
+
+	for(n = 0; n < unrolled; n += 8) {
+		in[0] = _mm_mul_ps(_mm_load_ps(&s[n]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s[n+4]), int_scale);
+		in[0] = _mm_add_ps(in[0], _mm_load_ps(&dither[n]));
+		in[1] = _mm_add_ps(in[1], _mm_load_ps(&dither[n+4]));
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
+		out[0] = _mm_packs_epi32(out[0], out[1]);
+		_mm_storeu_si128((__m128i*)(&d[n]), out[0]);
+	}
+	for(; n < n_samples; n++) {
+		in[0] = _mm_mul_ss(_mm_load_ss(&s[n]), int_scale);
+		in[0] = _mm_add_ss(in[0], _mm_load_ss(&dither[n]));
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		d[n] = _mm_cvtss_si32(in[0]);
+	}
+}
+
+void
+conv_f32d_to_s16d_dither_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
+		uint32_t n_samples)
+{
+	uint32_t i, k, chunk, n_channels = conv->n_channels;
+
+	update_dither_sse2(conv, SPA_MIN(n_samples, conv->dither_size));
+
+	for(i = 0; i < n_channels; i++) {
+		const float *s = src[i];
+		int16_t *d = dst[i];
+		for(k = 0; k < n_samples; k += chunk) {
+			chunk = SPA_MIN(n_samples - k, conv->dither_size);
+			conv_f32_to_s16_1_dither_sse2(conv, &d[k], &s[k], chunk);
+		}
+	}
+}
+
 void
 conv_f32d_to_s16_2_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 		uint32_t n_samples)
@@ -1205,8 +1366,9 @@ conv_f32d_to_s16_2_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 	uint32_t n, unrolled;
 	__m128 in[4];
 	__m128i out[4];
-	__m128 int_max = _mm_set1_ps(S16_MAX_F);
-        __m128 int_min = _mm_sub_ps(_mm_setzero_ps(), int_max);
+	__m128 int_scale = _mm_set1_ps(S16_SCALE);
+	__m128 int_max = _mm_set1_ps(S16_MAX);
+        __m128 int_min = _mm_set1_ps(S16_MIN);
 
 	if (SPA_IS_ALIGNED(s0, 16) &&
 	    SPA_IS_ALIGNED(s1, 16))
@@ -1215,15 +1377,15 @@ conv_f32d_to_s16_2_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 		unrolled = 0;
 
 	for(n = 0; n < unrolled; n += 8) {
-		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n+0]), int_max);
-		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n+0]), int_max);
-		in[2] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_max);
-		in[3] = _mm_mul_ps(_mm_load_ps(&s1[n+4]), int_max);
+		in[0] = _mm_mul_ps(_mm_load_ps(&s0[n+0]), int_scale);
+		in[1] = _mm_mul_ps(_mm_load_ps(&s1[n+0]), int_scale);
+		in[2] = _mm_mul_ps(_mm_load_ps(&s0[n+4]), int_scale);
+		in[3] = _mm_mul_ps(_mm_load_ps(&s1[n+4]), int_scale);
 
-		out[0] = _mm_cvtps_epi32(in[0]);
-		out[1] = _mm_cvtps_epi32(in[1]);
-		out[2] = _mm_cvtps_epi32(in[2]);
-		out[3] = _mm_cvtps_epi32(in[3]);
+		out[0] = _mm_cvttps_epi32(in[0]);
+		out[1] = _mm_cvttps_epi32(in[1]);
+		out[2] = _mm_cvttps_epi32(in[2]);
+		out[3] = _mm_cvttps_epi32(in[3]);
 
 		out[0] = _mm_packs_epi32(out[0], out[2]);
 		out[1] = _mm_packs_epi32(out[1], out[3]);
@@ -1237,10 +1399,10 @@ conv_f32d_to_s16_2_sse2(struct convert *conv, void * SPA_RESTRICT dst[], const v
 		d += 16;
 	}
 	for(; n < n_samples; n++) {
-		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_max);
-		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_max);
-		in[0] = _mm_min_ss(int_max, _mm_max_ss(in[0], int_min));
-		in[1] = _mm_min_ss(int_max, _mm_max_ss(in[1], int_min));
+		in[0] = _mm_mul_ss(_mm_load_ss(&s0[n]), int_scale);
+		in[1] = _mm_mul_ss(_mm_load_ss(&s1[n]), int_scale);
+		in[0] = _MM_CLAMP_SS(in[0], int_min, int_max);
+		in[1] = _MM_CLAMP_SS(in[1], int_min, int_max);
 		d[0] = _mm_cvtss_si32(in[0]);
 		d[1] = _mm_cvtss_si32(in[1]);
 		d += 2;
