@@ -53,7 +53,9 @@
 #include "defs.h"
 #include "a2dp-codecs.h"
 
-#define NAME  "bluez5-device"
+static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.device");
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT &log_topic
 
 #define MAX_DEVICES	64
 
@@ -135,6 +137,7 @@ struct impl {
 
 	uint32_t profile;
 	unsigned int switching_codec:1;
+	unsigned int save_profile:1;
 	uint32_t prev_bt_connected_profiles;
 
 	const struct a2dp_codec **supported_codecs;
@@ -181,13 +184,17 @@ static void get_a2dp_codecs(struct impl *this, enum spa_bluetooth_audio_codec id
 	*codecs = NULL;
 }
 
-static const struct a2dp_codec *get_supported_a2dp_codec(struct impl *this, enum spa_bluetooth_audio_codec id)
+static const struct a2dp_codec *get_supported_a2dp_codec(struct impl *this, enum spa_bluetooth_audio_codec id, size_t *idx)
 {
 	const struct a2dp_codec *a2dp_codec = NULL;
 	size_t i;
-	for (i = 0; i < this->supported_codec_count; ++i)
-		if (this->supported_codecs[i]->id == id)
+	for (i = 0; i < this->supported_codec_count; ++i) {
+		if (this->supported_codecs[i]->id == id) {
 			a2dp_codec = this->supported_codecs[i];
+			if (idx)
+				*idx = i;
+		}
+	}
 	return a2dp_codec;
 }
 
@@ -330,35 +337,57 @@ static void node_update_soft_volumes(struct node *node, float hw_volume)
 	}
 }
 
-static void volume_changed(void *userdata)
+static bool node_update_volume_from_transport(struct node *node, bool reset)
 {
-	struct node *node = userdata;
 	struct impl *impl = node->impl;
 	struct spa_bt_transport_volume *t_volume;
 	float prev_hw_volume;
 
 	if (!node->transport || !spa_bt_transport_volume_enabled(node->transport))
-		return;
+		return false;
 
 	/* PW is the controller for remote device. */
 	if (impl->profile != DEVICE_PROFILE_A2DP
 	    && impl->profile !=  DEVICE_PROFILE_HSP_HFP)
-		return;
+		return false;
 
 	t_volume = &node->transport->volumes[node->id];
 
 	if (!t_volume->active)
-		return;
+		return false;
 
 	prev_hw_volume = node_get_hw_volume(node);
-	for (uint32_t i = 0; i < node->n_channels; ++i) {
-		node->volumes[i] = prev_hw_volume > 0.0f
-			? node->volumes[i] * t_volume->volume / prev_hw_volume
-			: t_volume->volume;
+
+	if (!reset) {
+		for (uint32_t i = 0; i < node->n_channels; ++i) {
+			node->volumes[i] = prev_hw_volume > 0.0f
+				? node->volumes[i] * t_volume->volume / prev_hw_volume
+				: t_volume->volume;
+		}
+	} else {
+		for (uint32_t i = 0; i < node->n_channels; ++i)
+			node->volumes[i] = t_volume->volume;
 	}
 
 	node_update_soft_volumes(node, t_volume->volume);
 
+	/*
+	 * Consider volume changes from the headset as requested
+	 * by the user, and to be saved by the SM.
+	 */
+	node->save = true;
+
+	return true;
+}
+
+static void volume_changed(void *userdata)
+{
+	struct node *node = userdata;
+	struct impl *impl = node->impl;
+
+	if (!node_update_volume_from_transport(node, false))
+		return;
+
 	emit_volume(impl, node);
 
 	impl->info.change_mask |= SPA_DEVICE_CHANGE_MASK_PARAMS;
@@ -462,6 +491,8 @@ static void emit_node(struct impl *this, struct spa_bt_transport *t,
 				this->nodes[id].volumes[i] = this->nodes[id].volumes[i % prev_channels];
 		}
 
+		node_update_volume_from_transport(&this->nodes[id], true);
+
 		boost = get_soft_volume_boost(&this->nodes[id]);
 		if (boost != 1.0f) {
 			size_t i;
@@ -512,11 +543,13 @@ static void dynamic_node_transport_state_changed(void *data,
 	if (state >= SPA_BT_TRANSPORT_STATE_PENDING && old < SPA_BT_TRANSPORT_STATE_PENDING) {
 		if (!SPA_FLAG_IS_SET(this->id, DYNAMIC_NODE_ID_FLAG)) {
 			SPA_FLAG_SET(this->id, DYNAMIC_NODE_ID_FLAG);
+			spa_bt_transport_keepalive(t, true);
 			emit_node(impl, t, this->id, this->factory_name, this->a2dp_duplex);
 		}
 	} else if (state < SPA_BT_TRANSPORT_STATE_PENDING && old >= SPA_BT_TRANSPORT_STATE_PENDING) {
 		if (SPA_FLAG_IS_SET(this->id, DYNAMIC_NODE_ID_FLAG)) {
 			SPA_FLAG_CLEAR(this->id, DYNAMIC_NODE_ID_FLAG);
+			spa_bt_transport_keepalive(t, false);
 			spa_device_emit_object_info(&impl->hooks, this->id, NULL);
 		}
 	}
@@ -579,7 +612,7 @@ static const struct spa_bt_transport_events dynamic_node_transport_events = {
 static void emit_dynamic_node(struct dynamic_node *this, struct impl *impl,
 	struct spa_bt_transport *t, uint32_t id, const char *factory_name, bool a2dp_duplex)
 {
-	spa_log_debug(impl->log, NAME": dynamic node, transport: %p->%p id: %08x->%08x",
+	spa_log_debug(impl->log, "dynamic node, transport: %p->%p id: %08x->%08x",
 		this->transport, t, this->id, id);
 
 	if (this->transport) {
@@ -681,6 +714,9 @@ static int emit_nodes(struct impl *this)
 				}
 			}
 		}
+
+		if (get_supported_a2dp_codec(this, this->props.codec, NULL) == NULL)
+			this->props.codec = 0;
 		break;
 	case DEVICE_PROFILE_HSP_HFP:
 		if (this->bt_dev->connected_profiles & SPA_BT_PROFILE_HEADSET_HEAD_UNIT) {
@@ -696,6 +732,9 @@ static int emit_nodes(struct impl *this)
 				emit_node(this, t, DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_SCO_SINK, false);
 			}
 		}
+
+		if (spa_bt_device_supports_hfp_codec(this->bt_dev, get_hfp_codec(this->props.codec)) != 1)
+			this->props.codec = 0;
 		break;
 	default:
 		return -EINVAL;
@@ -745,15 +784,17 @@ static void emit_remove_nodes(struct impl *this)
 static bool validate_profile(struct impl *this, uint32_t profile,
 		enum spa_bluetooth_audio_codec codec);
 
-static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_audio_codec codec)
+static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_audio_codec codec, bool save)
 {
 	if (!validate_profile(this, profile, codec)) {
-		spa_log_warn(this->log, NAME ": trying to set invalid profile %d, codec %d, %08x %08x",
+		spa_log_warn(this->log, "trying to set invalid profile %d, codec %d, %08x %08x",
 			    profile, codec,
 			    this->bt_dev->profiles, this->bt_dev->connected_profiles);
 		return -EINVAL;
 	}
 
+	this->save_profile = save;
+
 	if (this->profile == profile &&
 	    (this->profile != DEVICE_PROFILE_A2DP || codec == this->props.codec) &&
 	    (this->profile != DEVICE_PROFILE_HSP_HFP || codec == this->props.codec))
@@ -783,7 +824,7 @@ static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_a
 		ret = spa_bt_device_ensure_a2dp_codec(this->bt_dev, codecs);
 		if (ret < 0) {
 			if (ret != -ENOTSUP)
-				spa_log_error(this->log, NAME": failed to switch codec (%d), setting basic profile", ret);
+				spa_log_error(this->log, "failed to switch codec (%d), setting basic profile", ret);
 		} else {
 			return 0;
 		}
@@ -795,7 +836,7 @@ static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_a
 		ret = spa_bt_device_ensure_hfp_codec(this->bt_dev, get_hfp_codec(codec));
 		if (ret < 0) {
 			if (ret != -ENOTSUP)
-				spa_log_error(this->log, NAME": failed to switch codec (%d), setting basic profile", ret);
+				spa_log_error(this->log, "failed to switch codec (%d), setting basic profile", ret);
 		} else {
 			return 0;
 		}
@@ -820,13 +861,13 @@ static void codec_switched(void *userdata, int status)
 {
 	struct impl *this = userdata;
 
-	spa_log_debug(this->log, NAME": codec switched (status %d)", status);
+	spa_log_debug(this->log, "codec switched (status %d)", status);
 
 	this->switching_codec = false;
 
 	if (status < 0) {
 		/* Failed to switch: return to a fallback profile */
-		spa_log_error(this->log, NAME": failed to switch codec (%d), setting fallback profile", status);
+		spa_log_error(this->log, "failed to switch codec (%d), setting fallback profile", status);
 		if (this->profile == DEVICE_PROFILE_A2DP && this->props.codec != 0) {
 			this->props.codec = 0;
 		} else if (this->profile == DEVICE_PROFILE_HSP_HFP && this->props.codec != 0) {
@@ -859,7 +900,7 @@ static void profiles_changed(void *userdata, uint32_t prev_profiles, uint32_t pr
 	connected_change = (this->bt_dev->connected_profiles ^ prev_connected_profiles);
 
 	/* Profiles changed. We have to re-emit device information. */
-	spa_log_info(this->log, NAME": profiles changed to  %08x %08x (prev %08x %08x, change %08x)"
+	spa_log_info(this->log, "profiles changed to  %08x %08x (prev %08x %08x, change %08x)"
 		     " switching_codec:%d",
 		     this->bt_dev->profiles, this->bt_dev->connected_profiles,
 		     prev_profiles, prev_connected_profiles, connected_change,
@@ -882,22 +923,22 @@ static void profiles_changed(void *userdata, uint32_t prev_profiles, uint32_t pr
 	case DEVICE_PROFILE_AG:
 		nodes_changed = (connected_change & (SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY |
 						     SPA_BT_PROFILE_A2DP_SOURCE));
-		spa_log_debug(this->log, NAME": profiles changed: AG nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: AG nodes changed: %d",
 			      nodes_changed);
 		break;
 	case DEVICE_PROFILE_A2DP:
-		if (get_supported_a2dp_codec(this, this->props.codec) == NULL)
+		if (get_supported_a2dp_codec(this, this->props.codec, NULL) == NULL)
 			this->props.codec = 0;
 		nodes_changed = (connected_change & (SPA_BT_PROFILE_A2DP_SINK |
 						     SPA_BT_PROFILE_A2DP_SOURCE));
-		spa_log_debug(this->log, NAME": profiles changed: A2DP nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: A2DP nodes changed: %d",
 			      nodes_changed);
 		break;
 	case DEVICE_PROFILE_HSP_HFP:
 		if (spa_bt_device_supports_hfp_codec(this->bt_dev, get_hfp_codec(this->props.codec)) != 1)
 			this->props.codec = 0;
 		nodes_changed = (connected_change & SPA_BT_PROFILE_HEADSET_HEAD_UNIT);
-		spa_log_debug(this->log, NAME": profiles changed: HSP/HFP nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: HSP/HFP nodes changed: %d",
 			      nodes_changed);
 		break;
 	}
@@ -982,7 +1023,7 @@ static uint32_t profile_direction_mask(struct impl *this, uint32_t index, enum s
 		if (device->connected_profiles & SPA_BT_PROFILE_A2DP_SINK)
 			have_output = true;
 
-		a2dp_codec = get_supported_a2dp_codec(this, codec);
+		a2dp_codec = get_supported_a2dp_codec(this, codec, NULL);
 		if (a2dp_codec && a2dp_codec->duplex_codec)
 			have_input = true;
 		break;
@@ -1053,7 +1094,8 @@ static uint32_t get_index_from_profile(struct impl *this, uint32_t profile, enum
 	return SPA_ID_INVALID;
 }
 
-static bool find_hsp_hfp_profile(struct impl *this) {
+static bool set_initial_hsp_hfp_profile(struct impl *this)
+{
 	struct spa_bt_transport *t;
 	int i;
 
@@ -1066,6 +1108,9 @@ static bool find_hsp_hfp_profile(struct impl *this) {
 			this->profile = (i & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY) ?
 				DEVICE_PROFILE_AG : DEVICE_PROFILE_HSP_HFP;
 			this->props.codec = get_hfp_codec_id(t->codec);
+
+			spa_log_debug(this->log, "initial profile HSP/HFP profile:%d codec:%d",
+					this->profile, this->props.codec);
 			return true;
 		}
 	}
@@ -1077,6 +1122,8 @@ static void set_initial_profile(struct impl *this)
 	struct spa_bt_transport *t;
 	int i;
 
+	this->switching_codec = false;
+
 	if (this->supported_codecs)
 		free(this->supported_codecs);
 	this->supported_codecs = spa_bt_device_get_supported_a2dp_codecs(
@@ -1085,11 +1132,12 @@ static void set_initial_profile(struct impl *this)
 	/* Prefer A2DP, then HFP, then null, but select AG if the device
 	   appears not to have A2DP_SINK or any HEAD_UNIT profile */
 
-	// If default profile is set to HSP/HFP, first try those and exit if found
-	const char *str;
+	/* If default profile is set to HSP/HFP, first try those and exit if found. */
 	if (this->bt_dev->settings != NULL) {
-		str = spa_dict_lookup(this->bt_dev->settings, "device.profile");
-		if (str != NULL && spa_streq(str, "headset-head-unit") && find_hsp_hfp_profile(this))
+		const char *str = spa_dict_lookup(this->bt_dev->settings, "bluez5.profile");
+		if (spa_streq(str, "off"))
+			goto off;
+		if (spa_streq(str, "headset-head-unit") && set_initial_hsp_hfp_profile(this))
 			return;
 	}
 
@@ -1102,18 +1150,25 @@ static void set_initial_profile(struct impl *this)
 			this->profile = (i == SPA_BT_PROFILE_A2DP_SOURCE) ?
 				DEVICE_PROFILE_AG : DEVICE_PROFILE_A2DP;
 			this->props.codec = t->a2dp_codec->id;
+			spa_log_debug(this->log, "initial profile A2DP profile:%d codec:%d",
+					this->profile, this->props.codec);
 			return;
 		}
 	}
 
-	if (find_hsp_hfp_profile(this)) return;
+	if (set_initial_hsp_hfp_profile(this))
+		return;
+
+off:
+	spa_log_debug(this->log, "initial profile off");
 
 	this->profile = DEVICE_PROFILE_OFF;
 	this->props.codec = 0;
 }
 
 static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *b,
-		uint32_t id, uint32_t index, uint32_t profile_index, enum spa_bluetooth_audio_codec codec)
+		uint32_t id, uint32_t index, uint32_t profile_index, enum spa_bluetooth_audio_codec codec,
+		bool current)
 {
 	struct spa_bt_device *device = this->bt_dev;
 	struct spa_pod_frame f[2];
@@ -1122,11 +1177,13 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 	char *desc_and_codec = NULL;
 	uint32_t n_source = 0, n_sink = 0;
 	uint32_t capture[1] = { DEVICE_ID_SOURCE }, playback[1] = { DEVICE_ID_SINK };
+	int priority;
 
 	switch (profile_index) {
 	case DEVICE_PROFILE_OFF:
 		name = "off";
 		desc = _("Off");
+		priority = 0;
 		break;
 	case DEVICE_PROFILE_AG:
 	{
@@ -1138,6 +1195,7 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 			name = "audio-gateway";
 			desc = _("Audio Gateway (A2DP Source & HSP/HFP AG)");
 		}
+		priority = 256;
 		break;
 	}
 	case DEVICE_PROFILE_A2DP:
@@ -1151,7 +1209,8 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		name = spa_bt_profile_name(profile);
 		n_sink++;
 		if (codec) {
-			const struct a2dp_codec *a2dp_codec = get_supported_a2dp_codec(this, codec);
+			size_t idx;
+			const struct a2dp_codec *a2dp_codec = get_supported_a2dp_codec(this, codec, &idx);
 			if (a2dp_codec == NULL) {
 				errno = EINVAL;
 				return NULL;
@@ -1167,12 +1226,14 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 
 			}
 			desc = desc_and_codec;
+			priority = 16 + this->supported_codec_count - idx;  /* order as in codec list */
 		} else {
 			if (profile == SPA_BT_PROFILE_A2DP_SINK) {
 				desc = _("High Fidelity Playback (A2DP Sink)");
 			} else {
 				desc = _("High Fidelity Duplex (A2DP Source/Sink)");
 			}
+			priority = 16;
 		}
 		break;
 	}
@@ -1201,8 +1262,10 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 			desc_and_codec = spa_aprintf(_("Headset Head Unit (HSP/HFP, codec %s)"),
 						get_hfp_codec_description(hfp_codec));
 			desc = desc_and_codec;
+			priority = 1 + hfp_codec;  /* prefer msbc over cvsd */
 		} else {
 			desc = _("Headset Head Unit (HSP/HFP)");
+			priority = 1;
 		}
 		break;
 	}
@@ -1217,6 +1280,7 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		SPA_PARAM_PROFILE_name, SPA_POD_String(name),
 		SPA_PARAM_PROFILE_description, SPA_POD_String(desc),
 		SPA_PARAM_PROFILE_available, SPA_POD_Id(SPA_PARAM_AVAILABILITY_yes),
+		SPA_PARAM_PROFILE_priority, SPA_POD_Int(priority),
 		0);
 	if (n_source > 0 || n_sink > 0) {
 		spa_pod_builder_prop(b, SPA_PARAM_PROFILE_classes, 0);
@@ -1237,6 +1301,10 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		}
 		spa_pod_builder_pop(b, &f[1]);
 	}
+	if (current) {
+                spa_pod_builder_prop(b, SPA_PARAM_PROFILE_save, 0);
+                spa_pod_builder_bool(b, this->save_profile);
+	}
 
 	if (name_and_codec)
 		free(name_and_codec);
@@ -1253,20 +1321,21 @@ static bool validate_profile(struct impl *this, uint32_t profile,
 	uint8_t buffer[1024];
 
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
-	return (build_profile(this, &b, 0, 0, profile, codec) != NULL);
+	return (build_profile(this, &b, 0, 0, profile, codec, false) != NULL);
 }
 
 static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
-		uint32_t id, uint32_t port, uint32_t dev, uint32_t profile)
+		uint32_t id, uint32_t port, uint32_t profile)
 {
 	struct spa_bt_device *device = this->bt_dev;
 	struct spa_pod_frame f[2];
 	enum spa_direction direction;
-	const char *name_prefix, *description, *port_type;
+	const char *name_prefix, *description, *hfp_description, *port_type;
 	enum spa_bt_form_factor ff;
 	enum spa_bluetooth_audio_codec codec;
 	char name[128];
 	uint32_t i, j, mask, next;
+	uint32_t dev = SPA_ID_INVALID, enum_dev;
 
 	ff = spa_bt_form_factor_from_class(device->bluetooth_class);
 
@@ -1274,52 +1343,62 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 	case SPA_BT_FORM_FACTOR_HEADSET:
 		name_prefix = "headset";
 		description = _("Headset");
+		hfp_description = _("Handsfree");
 		port_type = "headset";
 		break;
 	case SPA_BT_FORM_FACTOR_HANDSFREE:
 		name_prefix = "handsfree";
 		description = _("Handsfree");
+		hfp_description = _("Handsfree (HFP)");
 		port_type = "handsfree";
 		break;
 	case SPA_BT_FORM_FACTOR_MICROPHONE:
 		name_prefix = "microphone";
 		description = _("Microphone");
+		hfp_description = _("Handsfree");
 		port_type = "mic";
 		break;
 	case SPA_BT_FORM_FACTOR_SPEAKER:
 		name_prefix = "speaker";
 		description = _("Speaker");
+		hfp_description = _("Handsfree");
 		port_type = "speaker";
 		break;
 	case SPA_BT_FORM_FACTOR_HEADPHONE:
 		name_prefix = "headphone";
 		description = _("Headphone");
+		hfp_description = _("Handsfree");
 		port_type = "headphones";
 		break;
 	case SPA_BT_FORM_FACTOR_PORTABLE:
 		name_prefix = "portable";
 		description = _("Portable");
+		hfp_description = _("Handsfree");
 		port_type = "portable";
 		break;
 	case SPA_BT_FORM_FACTOR_CAR:
 		name_prefix = "car";
 		description = _("Car");
+		hfp_description = _("Handsfree");
 		port_type = "car";
 		break;
 	case SPA_BT_FORM_FACTOR_HIFI:
 		name_prefix = "hifi";
 		description = _("HiFi");
+		hfp_description = _("Handsfree");
 		port_type = "hifi";
 		break;
 	case SPA_BT_FORM_FACTOR_PHONE:
 		name_prefix = "phone";
 		description = _("Phone");
+		hfp_description = _("Handsfree");
 		port_type = "phone";
 		break;
 	case SPA_BT_FORM_FACTOR_UNKNOWN:
 	default:
 		name_prefix = "bluetooth";
 		description = _("Bluetooth");
+		hfp_description = _("Bluetooth (HFP)");
 		port_type = "bluetooth";
 		break;
 	}
@@ -1328,16 +1407,51 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 	case 0:
 		direction = SPA_DIRECTION_INPUT;
 		snprintf(name, sizeof(name), "%s-input", name_prefix);
+		enum_dev = DEVICE_ID_SOURCE;
+		if (profile == DEVICE_PROFILE_A2DP)
+			dev = enum_dev;
+		else if (profile != SPA_ID_INVALID)
+			enum_dev = SPA_ID_INVALID;
 		break;
 	case 1:
 		direction = SPA_DIRECTION_OUTPUT;
 		snprintf(name, sizeof(name), "%s-output", name_prefix);
+		enum_dev = DEVICE_ID_SINK;
+		if (profile == DEVICE_PROFILE_A2DP)
+			dev = enum_dev;
+		else if (profile != SPA_ID_INVALID)
+			enum_dev = SPA_ID_INVALID;
+		break;
+	case 2:
+		direction = SPA_DIRECTION_INPUT;
+		snprintf(name, sizeof(name), "%s-hf-input", name_prefix);
+		description = hfp_description;
+		enum_dev = DEVICE_ID_SOURCE;
+		if (profile == DEVICE_PROFILE_HSP_HFP)
+			dev = enum_dev;
+		else if (profile != SPA_ID_INVALID)
+			enum_dev = SPA_ID_INVALID;
+		break;
+	case 3:
+		direction = SPA_DIRECTION_OUTPUT;
+		snprintf(name, sizeof(name), "%s-hf-output", name_prefix);
+		description = hfp_description;
+		enum_dev = DEVICE_ID_SINK;
+		if (profile == DEVICE_PROFILE_HSP_HFP)
+			dev = enum_dev;
+		else if (profile != SPA_ID_INVALID)
+			enum_dev = SPA_ID_INVALID;
 		break;
 	default:
 		errno = EINVAL;
 		return NULL;
 	}
 
+	if (enum_dev == SPA_ID_INVALID) {
+		errno = EINVAL;
+		return NULL;
+	}
+
 	spa_pod_builder_push_object(b, &f[0], SPA_TYPE_OBJECT_ParamRoute, id);
 	spa_pod_builder_add(b,
 		SPA_PARAM_ROUTE_index, SPA_POD_Int(port),
@@ -1362,6 +1476,11 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 	for (i = 1; (j = get_profile_from_index(this, i, &next, &codec)) != SPA_ID_INVALID; i = next) {
 		uint32_t profile_mask;
 
+		if (j == DEVICE_PROFILE_A2DP && !(port == 0 || port == 1))
+			continue;
+		if (j == DEVICE_PROFILE_HSP_HFP && !(port == 2 || port == 3))
+			continue;
+
 		profile_mask = profile_direction_mask(this, j, codec);
 		if (!(profile_mask & (1 << direction)))
 			continue;
@@ -1428,8 +1547,7 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 
 	spa_pod_builder_prop(b, SPA_PARAM_ROUTE_devices, 0);
 	spa_pod_builder_push_array(b, &f[1]);
-	/* port and device indexes are the same, 0=source, 1=sink */
-	spa_pod_builder_int(b, port);
+	spa_pod_builder_int(b, enum_dev);
 	spa_pod_builder_pop(b, &f[1]);
 
 	if (profile != SPA_ID_INVALID) {
@@ -1481,7 +1599,7 @@ static struct spa_pod *build_prop_info(struct impl *this, struct spa_pod_builder
 	/* Transport codec */
 	spa_pod_builder_prop(b, SPA_PROP_INFO_id, 0);
 	spa_pod_builder_id(b, SPA_PROP_bluetoothAudioCodec);
-	spa_pod_builder_prop(b, SPA_PROP_INFO_name, 0);
+	spa_pod_builder_prop(b, SPA_PROP_INFO_description, 0);
 	spa_pod_builder_string(b, "Air codec");
 	spa_pod_builder_prop(b, SPA_PROP_INFO_type, 0);
 	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_Enum, 0);
@@ -1568,7 +1686,7 @@ static int impl_enum_params(void *object, int seq,
 		case DEVICE_PROFILE_AG:
 		case DEVICE_PROFILE_A2DP:
 		case DEVICE_PROFILE_HSP_HFP:
-			param = build_profile(this, &b, id, result.index, profile, codec);
+			param = build_profile(this, &b, id, result.index, profile, codec, false);
 			if (param == NULL)
 				goto next;
 			break;
@@ -1584,7 +1702,7 @@ static int impl_enum_params(void *object, int seq,
 		switch (result.index) {
 		case 0:
 			index = get_index_from_profile(this, this->profile, this->props.codec);
-			param = build_profile(this, &b, id, index, this->profile, this->props.codec);
+			param = build_profile(this, &b, id, index, this->profile, this->props.codec, true);
 			if (param == NULL)
 				return 0;
 			break;
@@ -1596,9 +1714,8 @@ static int impl_enum_params(void *object, int seq,
 	case SPA_PARAM_EnumRoute:
 	{
 		switch (result.index) {
-		case 0: case 1:
-			param = build_route(this, &b, id, result.index,
-					SPA_ID_INVALID, SPA_ID_INVALID);
+		case 0: case 1: case 2: case 3:
+			param = build_route(this, &b, id, result.index, SPA_ID_INVALID);
 			if (param == NULL)
 				goto next;
 			break;
@@ -1610,9 +1727,8 @@ static int impl_enum_params(void *object, int seq,
 	case SPA_PARAM_Route:
 	{
 		switch (result.index) {
-		case 0: case 1:
-			param = build_route(this, &b, id, result.index,
-					result.index, this->profile);
+		case 0: case 1: case 2: case 3:
+			param = build_route(this, &b, id, result.index, this->profile);
 			if (param == NULL)
 				goto next;
 			break;
@@ -1829,13 +1945,15 @@ static int impl_set_param(void *object,
 		uint32_t idx, next;
 		uint32_t profile;
 		enum spa_bluetooth_audio_codec codec;
+		bool save = false;
 
 		if (param == NULL)
 			return -EINVAL;
 
 		if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamProfile, NULL,
-				SPA_PARAM_PROFILE_index, SPA_POD_Int(&idx))) < 0) {
+				SPA_PARAM_PROFILE_index, SPA_POD_Int(&idx),
+				SPA_PARAM_PROFILE_save, SPA_POD_OPT_Bool(&save))) < 0) {
 			spa_log_warn(this->log, "can't parse profile");
 			spa_debug_pod(0, NULL, param);
 			return res;
@@ -1845,8 +1963,8 @@ static int impl_set_param(void *object,
 		if (profile == SPA_ID_INVALID)
 			return -EINVAL;
 
-		spa_log_debug(this->log, NAME": setting profile %d codec:%d", profile, codec);
-		return set_profile(this, profile, codec);
+		spa_log_debug(this->log, "setting profile %d codec:%d save:%d", profile, codec, (int)save);
+		return set_profile(this, profile, codec, save);
 	}
 	case SPA_PARAM_Route:
 	{
@@ -1898,6 +2016,8 @@ static int impl_set_param(void *object,
 			return res;
 		}
 
+		spa_log_debug(this->log, "setting props codec:%d", codec_id);
+
 		if (codec_id == SPA_ID_INVALID)
 			return 0;
 
@@ -1905,16 +2025,16 @@ static int impl_set_param(void *object,
 			size_t j;
 			for (j = 0; j < this->supported_codec_count; ++j) {
 				if (this->supported_codecs[j]->id == codec_id) {
-					return set_profile(this, this->profile, codec_id);
+					return set_profile(this, this->profile, codec_id, true);
 				}
 			}
 		} else if (this->profile == DEVICE_PROFILE_HSP_HFP) {
 			if (codec_id == SPA_BLUETOOTH_AUDIO_CODEC_CVSD &&
 					spa_bt_device_supports_hfp_codec(this->bt_dev, HFP_AUDIO_CODEC_CVSD) == 1) {
-				return set_profile(this, this->profile, codec_id);
+				return set_profile(this, this->profile, codec_id, true);
 			} else if (codec_id == SPA_BLUETOOTH_AUDIO_CODEC_MSBC &&
 					spa_bt_device_supports_hfp_codec(this->bt_dev, HFP_AUDIO_CODEC_MSBC) == 1) {
-				return set_profile(this, this->profile, codec_id);
+				return set_profile(this, this->profile, codec_id, true);
 			}
 		}
 		return -EINVAL;
@@ -2019,6 +2139,8 @@ impl_init(const struct spa_handle_factory *factory,
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
 	_i18n = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_I18N);
 
+	spa_log_topic_init(this->log, &log_topic);
+
 	if (info && (str = spa_dict_lookup(info, SPA_KEY_API_BLUEZ5_DEVICE)))
 		sscanf(str, "pointer:%p", &this->bt_dev);
 
