@@ -55,13 +55,342 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 /**
  * \page page_module_filter_chain PipeWire Module: Filter-Chain
  *
+ * The filter-chain allows you to create an arbitrary processing graph
+ * from LADSPA, LV2 and builtin filters. This filter can be made into a
+ * virtual sink/source or between any 2 nodes in the graph.
+ *
+ * The filter chain is built with 2 streams, a capture stream providing
+ * the input to the filter chain and a playback stream sending out the
+ * filtered stream to the next nodes in the graph.
+ *
+ * Because both ends of the filter-chain are built with streams, the session
+ * manager can manage the configuration and connection with the sinks and
+ * sources automatically.
+ *
+ * ## Module Options
+ *
+ * - `node.description`: a human readable name for the filter chain
+ * - `filter.graph = []`: a description of the filter graph to run, see below
+ * - `capture.props = {}`: properties to be passed to the input stream
+ * - `playback.props = {}`: properties to be passed to the output stream
+ *
+ * ## Filter graph description
+ *
+ * The general structure of the graph description is as follows:
+ *
+ *\code{.unparsed}
+ *     filter.graph = {
+ *         nodes = [
+ *             {
+ *                 type = <ladspa | lv2 | builtin>
+ *                 name = <name>
+ *                 plugin = <plugin>
+ *                 label = <label>
+ *                 config = {
+ *                     <configkey> = <value> ...
+ *                 }
+ *                 control = {
+ *                     <controlname|controlindex> = <value> ...
+ *                 }
+ *             }
+ *             ...
+ *         ]
+ *         links = [
+ *             { output = <portname> input = <portname> }
+ *             ...
+ *         ]
+ *         inputs = [ <portname> ... ]
+ *         outputs = [ <portname> ... ]
+ *    }
+ *\endcode
+ *
+ * ### Nodes
+ *
+ * Nodes describe the processing filters in the graph. Use a tool like lv2ls
+ * or listplugins to get a list of available plugins, labels and the port names.
+ *
+ * - `type` is one of `ladspa`, `lv2` or `builtin`
+ * - `name` is the name for this node, you might need this later to refer to this node
+ *    and its ports when setting controls or making links.
+ * - `plugin` is the type specific plugin name.
+ *    - For LADSPA plugins it will append `.so` to find the shared object with that
+ *       name in the LADSPA plugin path.
+ *    - For LV2, this is the plugin URI obtained with lv2ls.
+ *    - For builtin this is ignored
+ * - `label` is the type specific filter inside the plugin.
+ *    - For LADSPA this is the label
+ *    - For LV2 this is unused
+ *    - For builtin this is the name of the filter to use
+ *
+ * - `config` contains a filter specific configuration section. The convolver
+ *            plugin needs this.
+ * - `control` contains the initial values for the control ports of the filter.
+ *
+ * ### Links
+ *
+ * Links can be made between ports of nodes. The `portname` is given as
+ * `<node_name>:<port_name>`.
+ *
+ * You can tee the output of filters to multiple other filters. You need to
+ * use a mixer if you want the output of multiple filters to go into one
+ * filter input port.
+ *
+ * links can be omited when the graph has just 1 filter.
+ *
+ * ### Inputs and Outputs
+ *
+ * These are the entry and exit ports into the graph definition. Their number
+ * defines the number of channels used by the filter-chain.
+ *
+ * The `<portname>` can be `null` when a channel is to be ignored.
+ *
+ * Each input/output in the graph can only be linked to one filter input/output.
+ * You need to use the copy builtin filter if the stream signal needs to be routed
+ * to multiple filters. You need to use the mixer builtin plugin if multiple graph
+ * outputs need to go to one output stream.
+ *
+ * inputs and outputs can be omitted, in which case the filter-chain will use all
+ * inputs from the first filter and all outputs from the last filter node. The
+ * graph will then be duplicated as many times to match the number of input/output
+ * channels of the streams.
+ *
+ * ## Builtin filters
+ *
+ * There are some useful builtin filters available. You select them with the label
+ * of the filter node.
+ *
+ * ### Mixer
+ *
+ * Use the `mixer` plugin if you have multiple input signals that need to be mixed together.
+ *
+ * The mixer plugin has up to 8 input ports labeled "In 1" to "In 8" and each with
+ * a gain control labeled "Gain 1" to "Gain 8". There is an output port labeled
+ * "Out". Unused input ports will be ignoded and not cause overhead.
+ *
+ * ### Copy
+ *
+ * Use the `copy` plugin if you need to copy a stream input signal to multiple filters.
+ *
+ * It has one input port "In" and one output port "Out".
+ *
+ * ### Biquads
+ *
+ * Biquads can be used to do all kinds of filtering. They are also used when creating
+ * equalizers.
+ *
+ * All biquad filters have an input port "In" and an output port "Out". They have
+ * a "Freq", "Q" and "Gain" control. Their meaning depends on the particular biquad that
+ * is used. The following labels can be used:
+ *
+ * - `bq_lowpass` a lowpass filter.
+ * - `bq_highpass` a highpass filter.
+ * - `bq_bandpass` a bandpass filter.
+ * - `bq_lowshelf` a low shelf filter.
+ * - `bq_highshelf` a high shelf filter.
+ * - `bq_peaking` a peaking filter.
+ * - `bq_notch` a notch filter.
+ * - `bq_allpass` an allpass filter.
+ *
+ * ### Convolver
+ *
+ * The convolver can be used to apply an impulse response to a signal. It is usually used
+ * for reverbs or virtual surround. The convolver is implemented with a fast FFT
+ * implementation.
+ *
+ * The convolver has an input port "In" and an output port "Out". It requires a config
+ * section in the node declaration in this format:
+ *
+ *\code{.unparsed}
+ * filter.graph = {
+ *     nodes = [
+ *         {
+ *             type   = builtin
+ *             name   = ...
+ *             label  = convolver
+ *             config = {
+ *                 blocksize = ...
+ *                 tailsize = ...
+ *                 gain = ...
+ *                 delay = ...
+ *                 filename = ...
+ *                 offset = ...
+ *                 length = ...
+ *                 channel = ...
+ *             }
+ *             ...
+ *         }
+ *     }
+ *     ...
+ * }
+ *\endcode
+ *
+ * - `blocksize` specifies the size of the blocks to use in the FFT. It is a value
+ *               between 64 and 256. When not specified, this value is
+ *               computed automatically from the number of samples in the file.
+ * - `tailsize` specifies the size of the tail blocks to use in the FFT.
+ * - `gain`     the overall gain to apply to the IR file.
+ * - `delay`    The extra delay (in samples) to add to the IR.
+ * - `filename` The IR to load or create. Possible values are:
+ *     - `/hilbert` creates a [hilbert function](https://en.wikipedia.org/wiki/Hilbert_transform)
+ *                that can be used to phase shift the signal by +/-90 degrees. The
+ *                `length` will be used as the number of coefficients.
+ *     - `/dirac` creates a [Dirac function](https://en.wikipedia.org/wiki/Dirac_delta_function) that
+ *                 can be used as gain.
+ *     - A filename to load as the IR. This needs to be a file format supported
+ *               by sndfile.
+ * - `offset`  The sample offset in the file as the start of the IR.
+ * - `length`  The number of samples to use as the IR.
+ * - `channel` The channel to use from the file as the IR.
+ *
+ * ### Delay
+ *
+ * The delay can be used to delay a signal in time.
+ *
+ * The delay has an input port "In" and an output port "Out". It also has
+ * a "Delay (s)" control port. It requires a config section in the node declaration
+ * in this format:
+ *
+ *\code{.unparsed}
+ * filter.graph = {
+ *     nodes = [
+ *         {
+ *             type   = builtin
+ *             name   = ...
+ *             label  = delay
+ *             config = {
+ *                 "max-delay" = ...
+ *             }
+ *             control = {
+ *                 "Delay (s)" = ...
+ *             }
+ *             ...
+ *         }
+ *     }
+ *     ...
+ * }
+ *\endcode
+ *
+ * - `max-delay` the maximum delay in seconds. The "Delay (s)" parameter will
+ *              be clamped to this value.
+ *
+ * ## General options
+ *
+ * Options with well-known behavior. Most options can be added to the global
+ * configuration or the individual streams:
+ *
+ * - \ref PW_KEY_REMOTE_NAME
+ * - \ref PW_KEY_AUDIO_RATE
+ * - \ref PW_KEY_AUDIO_CHANNELS
+ * - \ref SPA_KEY_AUDIO_POSITION
+ * - \ref PW_KEY_MEDIA_NAME
+ * - \ref PW_KEY_NODE_LATENCY
+ * - \ref PW_KEY_NODE_DESCRIPTION
+ * - \ref PW_KEY_NODE_GROUP
+ * - \ref PW_KEY_NODE_LINK_GROUP
+ * - \ref PW_KEY_NODE_VIRTUAL
+ * - \ref PW_KEY_NODE_NAME: See notes below. If not specified, defaults to
+ *   	'filter-chain-<pid>-<module-id>'.
+ *
+ * Stream only properties:
+ *
+ * - \ref PW_KEY_MEDIA_CLASS
+ * - \ref PW_KEY_NODE_NAME:  if not given per stream, the global node.name will be
+ *         prefixed with 'input.' and 'output.' to generate a capture and playback
+ *         stream node.name respectively.
+ *
+ * ## Example configuration of a virtual source
+ *
+ * This example uses the rnnoise LADSPA plugin to create a new
+ * virtual source.
+ *
+ *\code{.unparsed}
+ * context.modules = [
+ * {   name = libpipewire-module-filter-chain
+ *     args = {
+ *         node.description =  "Noise Canceling source"
+ *         media.name =  "Noise Canceling source"
+ *         filter.graph = {
+ *             nodes = [
+ *                 {
+ *                     type = ladspa
+ *                     name = rnnoise
+ *                     plugin = ladspa/librnnoise_ladspa
+ *                     label = noise_suppressor_stereo
+ *                     control = {
+ *                         "VAD Threshold (%)" 50.0
+ *                     }
+ *                 }
+ *             ]
+ *         }
+ *         capture.props = {
+ *             node.name =  "capture.rnnoise_source"
+ *             node.passive = true
+ *         }
+ *         playback.props = {
+ *             node.name =  "rnnoise_source"
+ *             media.class = Audio/Source
+ *         }
+ *     }
+ * }
+ * ]
+ *\endcode
+ *
+ * ## Example configuration of a Dolby Surround encoder virtual Sink
+ *
+ * This example uses the ladpsa surround encoder to encode a 5.1 signal
+ * to a stereo Dolby Surround signal.
+ *
+ *\code{.unparsed}
+ *
+ *\code{.unparsed}
+ * context.modules = [
+ * {   name = libpipewire-module-filter-chain
+ *     args = {
+ *         node.description = "Dolby Surround Sink"
+ *         media.name       = "Dolby Surround Sink"
+ *         filter.graph = {
+ *             nodes = [
+ *                 {
+ *                     type  = builtin
+ *                     name  = mixer
+ *                     label = mixer
+ *                     control = { "Gain 1" = 0.5 "Gain 2" = 0.5 }
+ *                 }
+ *                 {
+ *                     type   = ladspa
+ *                     name   = enc
+ *                     plugin = surround_encoder_1401
+ *                     label  = surroundEncoder
+ *                 }
+ *             ]
+ *             links = [
+ *                 { output = "mixer:Out" input = "enc:S" }
+ *             ]
+ *             inputs  = [ "enc:L" "enc:R" "enc:C" null "mixer:In 1" "mixer:In 2" ]
+ *             outputs = [ "enc:Lt" "enc:Rt" ]
+ *         }
+ *         capture.props = {
+ *             node.name      = "effect_input.dolby_surround"
+ *             media.class    = Audio/Sink
+ *             audio.channels = 6
+ *             audio.position = [ FL FR FC LFE SL SR ]
+ *         }
+ *         playback.props = {
+ *             node.name      = "effect_output.dolby_surround"
+ *             node.passive   = true
+ *             audio.channels = 2
+ *             audio.position = [ FL FR ]
+ *         }
+ *     }
+ * }
+ * ]
+ *\endcode
  */
 static const struct spa_dict_item module_props[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Wim Taymans <wim.taymans@gmail.com>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "Create filter chain streams" },
 	{ PW_KEY_MODULE_USAGE, " [ remote.name=<remote> ] "
 				"[ node.latency=<latency as fraction> ] "
-				"[ node.name=<name of the nodes> ] "
 				"[ node.description=<description of the nodes> ] "
 				"[ audio.rate=<sample rate> ] "
 				"[ audio.channels=<number of channels> ] "
@@ -69,7 +398,7 @@ static const struct spa_dict_item module_props[] = {
 				"filter.graph = [ "
 				"    nodes = [ "
 				"        { "
-				"          type = ladspa "
+				"          type = <ladspa | lv2 | builtin> "
 				"          name = <name> "
 				"          plugin = <plugin> "
 				"          label = <label> "
@@ -77,7 +406,7 @@ static const struct spa_dict_item module_props[] = {
 				"             <configkey> = <value> ... "
 				"          } "
 				"          control = { "
-				"             <controlname> = <value> ... "
+				"             <controlname|controlindex> = <value> ... "
 				"          } "
 				"        } "
 				"    ] "
@@ -260,7 +589,7 @@ static void capture_process(void *d)
 	struct impl *impl = d;
 	struct pw_buffer *in, *out;
 	struct graph *graph = &impl->graph;
-	uint32_t i, size = 0, n_hndl = graph->n_hndl;
+	uint32_t i, outsize = 0, n_hndl = graph->n_hndl;
 	int32_t stride = 0;
 
 	if ((in = pw_stream_dequeue_buffer(impl->capture)) == NULL)
@@ -275,10 +604,16 @@ static void capture_process(void *d)
 	for (i = 0; i < in->buffer->n_datas; i++) {
 		struct spa_data *ds = &in->buffer->datas[i];
 		struct graph_port *port = &graph->input[i];
+		uint32_t offs, size;
+
+		offs = SPA_MIN(ds->chunk->offset, ds->maxsize);
+		size = SPA_MIN(ds->chunk->size, ds->maxsize - offs);
+
 		if (port->desc)
 			port->desc->connect_port(port->hndl, port->port,
-				SPA_MEMBER(ds->data, ds->chunk->offset, void));
-		size = SPA_MAX(size, ds->chunk->size);
+				SPA_PTROFF(ds->data, offs, void));
+
+		outsize = SPA_MAX(outsize, size);
 		stride = SPA_MAX(stride, ds->chunk->stride);
 	}
 	for (i = 0; i < out->buffer->n_datas; i++) {
@@ -287,14 +622,14 @@ static void capture_process(void *d)
 		if (port->desc)
 			port->desc->connect_port(port->hndl, port->port, dd->data);
 		else
-			memset(dd->data, 0, size);
+			memset(dd->data, 0, outsize);
 		dd->chunk->offset = 0;
-		dd->chunk->size = size;
+		dd->chunk->size = outsize;
 		dd->chunk->stride = stride;
 	}
 	for (i = 0; i < n_hndl; i++) {
 		struct graph_hndl *hndl = &graph->hndl[i];
-		hndl->desc->run(hndl->hndl, size / sizeof(float));
+		hndl->desc->run(hndl->hndl, outsize / sizeof(float));
 	}
 
 done:
@@ -322,6 +657,12 @@ static struct node *find_node(struct graph *graph, const char *name)
 	return NULL;
 }
 
+/* find a port by name. Valid syntax is:
+ *  "<node_name>:<port_name>"
+ *  "<node_name>:<port_id>"
+ *  "<port_name>"
+ *  "<port_id>"
+ *  When no node_name is given, the port is assumed in the current node.  */
 static struct port *find_port(struct node *node, const char *name, int descriptor)
 {
 	char *col, *node_name, *port_name, *str;
@@ -515,6 +856,7 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 	while (true) {
 		const char *name;
 		float value, *val = NULL;
+		double dbl_val;
 		bool bool_val;
 		int32_t int_val;
 
@@ -522,6 +864,9 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 			break;
 		if (spa_pod_parser_get_float(&prs, &value) >= 0) {
 			val = &value;
+		} else if (spa_pod_parser_get_double(&prs, &dbl_val) >= 0) {
+			value = dbl_val;
+			val = &value;
 		} else if (spa_pod_parser_get_int(&prs, &int_val) >= 0) {
 			value = int_val;
 			val = &value;
@@ -605,6 +950,13 @@ static void state_changed(void *data, enum pw_stream_state old,
 		pw_stream_flush(impl->capture, false);
 		graph_reset(graph);
 		break;
+	case PW_STREAM_STATE_UNCONNECTED:
+		pw_log_info("module %p: unconnected", impl);
+		pw_impl_module_schedule_destroy(impl->module);
+		break;
+	case PW_STREAM_STATE_ERROR:
+		pw_log_info("module %p: error: %s", impl, error);
+		break;
 	default:
 		break;
 	}
@@ -956,8 +1308,10 @@ static int parse_config(struct node *node, struct spa_json *config)
 	if (spa_json_is_container(val, len))
 		len = spa_json_container_len(config, val, len);
 
-	if ((node->config = malloc(len+1)) != NULL)
-		spa_json_parse_stringn(val, len, node->config, len+1);
+	if ((node->config = malloc(len+1)) == NULL)
+		return -errno;
+
+	spa_json_parse_stringn(val, len, node->config, len+1);
 
 	return 0;
 }
@@ -974,9 +1328,16 @@ static int parse_control(struct node *node, struct spa_json *control)
 
 	while (spa_json_get_string(control, key, sizeof(key)) > 0) {
 		float fl;
-		if (spa_json_get_float(control, &fl) <= 0)
+		const char *val;
+		int len;
+
+		if ((len = spa_json_next(control, &val)) < 0)
 			break;
-		set_control_value(node, key, &fl);
+
+		if (spa_json_parse_float(val, len, &fl) <= 0)
+			pw_log_warn("control '%s' expects a number, ignoring", key);
+		else
+			set_control_value(node, key, &fl);
 	}
 	return 0;
 }
@@ -996,6 +1357,11 @@ static int parse_link(struct graph *graph, struct spa_json *json)
 	struct port *in_port, *out_port;
 	struct link *link;
 
+	if (spa_list_is_empty(&graph->node_list)) {
+		pw_log_error("can't make links in graph without nodes");
+		return -EINVAL;
+	}
+
 	while (spa_json_get_string(json, key, sizeof(key)) > 0) {
 		if (spa_streq(key, "output")) {
 			if (spa_json_get_string(json, output, sizeof(output)) <= 0) {
@@ -1088,6 +1454,7 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	bool have_control = false;
 	bool have_config = false;
 	uint32_t i;
+	int res;
 
 	while (spa_json_get_string(json, key, sizeof(key)) > 0) {
 		if (spa_streq("type", key)) {
@@ -1119,6 +1486,8 @@ static int load_node(struct graph *graph, struct spa_json *json)
 		} else if (spa_streq("config", key)) {
 			config = SPA_JSON_SAVE(json);
 			have_config = true;
+			if (spa_json_next(json, &val) < 0)
+				break;
 		} else if (spa_json_next(json, &val) < 0)
 			break;
 	}
@@ -1180,7 +1549,8 @@ static int load_node(struct graph *graph, struct spa_json *json)
 		spa_list_init(&port->link_list);
 	}
 	if (have_config)
-		parse_config(node, &config);
+		if ((res = parse_config(node, &config)) < 0)
+			pw_log_warn("error parsing config: %s", spa_strerror(res));
 	if (have_control)
 		parse_control(node, &control);
 
@@ -1317,18 +1687,32 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		impl->playback_info.channels = n_output;
 
 	/* compare to the requested number of channels and duplicate the
-	 * graph m_hndl times when needed. */
+	 * graph n_hndl times when needed. */
 	n_hndl = impl->capture_info.channels / n_input;
 	if (n_hndl != impl->playback_info.channels / n_output) {
-		pw_log_error("invalid channels");
+		pw_log_error("invalid channels. The capture stream has %1$d channels and "
+				"the filter has %2$d inputs. The playback stream has %3$d channels "
+				"and the filter has %4$d outputs. capture:%1$d / input:%2$d != "
+				"playback:%3$d / output:%4$d. Check inputs and outputs objects.",
+				impl->capture_info.channels, n_input,
+				impl->playback_info.channels, n_output);
 		res = -EINVAL;
 		goto error;
 	}
 	if (n_hndl > MAX_HNDL) {
-		pw_log_error("too many channels");
+		pw_log_error("too many channels. %d > %d", n_hndl, MAX_HNDL);
 		res = -EINVAL;
 		goto error;
 	}
+	if (n_hndl == 0) {
+		n_hndl = 1;
+		pw_log_warn("The capture stream has %1$d channels and "
+				"the filter has %2$d inputs. The playback stream has %3$d channels "
+				"and the filter has %4$d outputs. Some filter ports will be "
+				"unconnected..",
+				impl->capture_info.channels, n_input,
+				impl->playback_info.channels, n_output);
+	}
 	pw_log_info("using %d instances %d %d", n_hndl, n_input, n_output);
 
 	/* now go over all nodes and create instances. */
@@ -1345,8 +1729,8 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		for (i = 0; i < n_hndl; i++) {
 			pw_log_info("instantiate %s %d", d->name, i);
 			if ((node->hndl[i] = d->instantiate(d, &impl->rate, i, node->config)) == NULL) {
-				pw_log_error("cannot create plugin instance");
-				res = -ENOMEM;
+				pw_log_error("cannot create plugin instance: %m");
+				res = -errno;
 				goto error;
 			}
 			node->n_hndl = i + 1;
@@ -1544,8 +1928,9 @@ error:
  */
 static int load_graph(struct graph *graph, struct pw_properties *props)
 {
-	struct spa_json it[4];
+	struct spa_json it[3];
 	struct spa_json inputs, outputs, *pinputs = NULL, *poutputs = NULL;
+	struct spa_json nodes, *pnodes = NULL, links, *plinks = NULL;
 	const char *json, *val;
 	char key[256];
 	int res;
@@ -1566,36 +1951,49 @@ static int load_graph(struct graph *graph, struct pw_properties *props)
 
 	while (spa_json_get_string(&it[1], key, sizeof(key)) > 0) {
 		if (spa_streq("nodes", key)) {
-			if (spa_json_enter_array(&it[1], &it[2]) <= 0) {
-				pw_log_error("nodes expect an array");
+			if (spa_json_enter_array(&it[1], &nodes) <= 0) {
+				pw_log_error("nodes expects an array");
 				return -EINVAL;
 			}
-			while (spa_json_enter_object(&it[2], &it[3]) > 0) {
-				if ((res = load_node(graph, &it[3])) < 0)
-					return res;
-			}
+			pnodes = &nodes;
 		}
 		else if (spa_streq("links", key)) {
-			if (spa_json_enter_array(&it[1], &it[2]) <= 0)
+			if (spa_json_enter_array(&it[1], &links) <= 0) {
+				pw_log_error("links expects an array");
 				return -EINVAL;
-
-			while (spa_json_enter_object(&it[2], &it[3]) > 0) {
-				if ((res = parse_link(graph, &it[3])) < 0)
-					return res;
 			}
+			plinks = &links;
 		}
 		else if (spa_streq("inputs", key)) {
-			if (spa_json_enter_array(&it[1], &inputs) <= 0)
+			if (spa_json_enter_array(&it[1], &inputs) <= 0) {
+				pw_log_error("inputs expects an array");
 				return -EINVAL;
+			}
 			pinputs = &inputs;
 		}
 		else if (spa_streq("outputs", key)) {
-			if (spa_json_enter_array(&it[1], &outputs) <= 0)
+			if (spa_json_enter_array(&it[1], &outputs) <= 0) {
+				pw_log_error("outputs expects an array");
 				return -EINVAL;
+			}
 			poutputs = &outputs;
 		} else if (spa_json_next(&it[1], &val) < 0)
 			break;
 	}
+	if (pnodes == NULL) {
+		pw_log_error("filter.graph is missing a nodes array");
+		return -EINVAL;
+	}
+	while (spa_json_enter_object(pnodes, &it[2]) > 0) {
+		if ((res = load_node(graph, &it[2])) < 0)
+			return res;
+	}
+	if (plinks != NULL) {
+		while (spa_json_enter_object(plinks, &it[2]) > 0) {
+			if ((res = parse_link(graph, &it[2])) < 0)
+				return res;
+		}
+	}
 	return setup_graph(graph, pinputs, poutputs);
 }
 
@@ -1699,10 +2097,9 @@ static void parse_audio_info(struct pw_properties *props, struct spa_audio_info_
 
 	*info = SPA_AUDIO_INFO_RAW_INIT(
 			.format = SPA_AUDIO_FORMAT_F32P);
-	if ((str = pw_properties_get(props, PW_KEY_AUDIO_RATE)) != NULL)
-		info->rate = atoi(str);
-	if ((str = pw_properties_get(props, PW_KEY_AUDIO_CHANNELS)) != NULL)
-		info->channels = atoi(str);
+	info->rate = pw_properties_get_int32(props, PW_KEY_AUDIO_RATE, info->rate);
+	info->channels = pw_properties_get_int32(props, PW_KEY_AUDIO_CHANNELS, info->channels);
+	info->channels = SPA_MIN(info->channels, SPA_AUDIO_MAX_CHANNELS);
 	if ((str = pw_properties_get(props, SPA_KEY_AUDIO_POSITION)) != NULL)
 		parse_position(info, str, strlen(str));
 }
@@ -1790,12 +2187,17 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	parse_audio_info(impl->capture_props, &impl->capture_info);
 	parse_audio_info(impl->playback_props, &impl->playback_info);
 
+	if ((str = pw_properties_get(props, PW_KEY_NODE_NAME)) == NULL) {
+		pw_properties_setf(props, PW_KEY_NODE_NAME,
+				"filter-chain-%u-%u", pid, id);
+		str = pw_properties_get(props, PW_KEY_NODE_NAME);
+	}
 	if (pw_properties_get(impl->capture_props, PW_KEY_NODE_NAME) == NULL)
 		pw_properties_setf(impl->capture_props, PW_KEY_NODE_NAME,
-				"input.filter-chain-%u-%u", pid, id);
+				"input.%s", str);
 	if (pw_properties_get(impl->playback_props, PW_KEY_NODE_NAME) == NULL)
 		pw_properties_setf(impl->playback_props, PW_KEY_NODE_NAME,
-				"output.filter-chain-%u-%u", pid, id);
+				"output.%s", str);
 
 	if (pw_properties_get(impl->capture_props, PW_KEY_MEDIA_NAME) == NULL)
 		pw_properties_setf(impl->capture_props, PW_KEY_MEDIA_NAME, "%s input",
